'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import random

from pm4py.objects.log.obj import EventStream, EventLog


def sample_stream(event_log, no_events=100):
    """
    Randomly sample a fixed number of events from the original event log

    Parameters
    -----------
    event_log
        Event log
    no_events
        Number of events that the sample should have

    Returns
    -----------
    newLog
        Filtered log
    """
    new_log = EventStream(attributes=event_log.attributes, extensions=event_log.extensions, globals=event_log.omni_present,
                          classifiers=event_log.classifiers)
    new_log._list = random.sample(event_log, min(no_events, len(event_log)))
    return new_log


def sample_log(log, no_traces=100):
    """
    Randomly sample a fixed number of traces from the original log

    Parameters
    -----------
    log
        Log
    no_traces
        Number of traces that the sample should have

    Returns
    -----------
    newLog
        Filtered log
    """
    new_log = EventLog(attributes=log.attributes, extensions=log.extensions, globals=log.omni_present,
                       classifiers=log.classifiers, properties=log.properties)
    new_log._list = random.sample(log, min(no_traces, len(log)))
    return new_log


def sample(log, n=100):
    """
    Randomly sample a fixed number of traces from the original log

    Parameters
    -----------
    log
        Trace/event log
    n
        Number of elements that the sample should have

    Returns
    -----------
    newLog
        Filtered log
    """

    if type(log) is EventLog:
        return sample_log(log, no_traces=n)

    return sample_stream(log, no_events=n)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import sys
from enum import Enum
from typing import Any, Optional, Dict, Union

import pandas as pd

from pm4py.util import constants, xes_constants, exec_utils
from copy import copy


class Parameters(Enum):
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY
    ATTRIBUTE_KEY = constants.PARAMETER_CONSTANT_ATTRIBUTE_KEY
    MIN_REP = "min_rep"
    MAX_REP = "max_rep"


def apply(df: pd.DataFrame, value: Any, parameters: Optional[Dict[Union[str, Parameters], Any]] = None) -> pd.DataFrame:
    """
    Filters the trace of the dataframe where the given attribute value is repeated
    (in a range of repetitions that is specified by the user)

    Parameters
    ----------------
    df
        Dataframe
    value
        Value that is investigated
    parameters
        Parameters of the filter, including:
        - Parameters.ATTRIBUTE_KEY => the attribute key
        - Parameters.MIN_REP => minimum number of repetitions
        - Parameters.MAX_REP => maximum number of repetitions
        - Parameters.CASE_ID_KEY => the columns of the dataframe that is the case identifier

    Returns
    ----------------
    filtered_df
        Filtered dataframe
    """
    if parameters is None:
        parameters = {}

    case_id_key = exec_utils.get_param_value(Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME)
    attribute_key = exec_utils.get_param_value(Parameters.ATTRIBUTE_KEY, parameters, xes_constants.DEFAULT_NAME_KEY)
    min_rep = exec_utils.get_param_value(Parameters.MIN_REP, parameters, 2)
    max_rep = exec_utils.get_param_value(Parameters.MAX_REP, parameters, sys.maxsize)

    filtered_df = df[df[attribute_key] == value]
    filtered_df = filtered_df.groupby(case_id_key).size().reset_index()
    filtered_df = filtered_df[filtered_df[0] >= min_rep]
    filtered_df = filtered_df[filtered_df[0] <= max_rep]

    ret = df[df[case_id_key].isin(filtered_df[case_id_key])]
    ret.attrs = copy(df.attrs) if hasattr(df, 'attrs') else {}
    return ret



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import numpy as np
from scipy.spatial.distance import squareform
from pm4py.algo.clustering.trace_attribute_driven.leven_dist import leven_dist_calc
from pm4py.algo.clustering.trace_attribute_driven.merge_log import merge_log
from pm4py.algo.clustering.trace_attribute_driven.dfg import dfg_dist
from pm4py.algo.clustering.trace_attribute_driven.variants import act_dist_calc
from pm4py.algo.clustering.trace_attribute_driven.variants import suc_dist_calc


def linkage_dfg_update(loglist, dist_mat, alpha, percent):
    index_list = []
    for i in range(len(dist_mat)):
        for j in range(i + 1, len(dist_mat)):
            index_list.append([i, j])

    y = squareform(dist_mat)
    n = len(dist_mat)  # The number of observations.
    Z = []
    cluster_size = dict(zip(range(n), np.ones(n)))  # record merged cluster size every step
    k = 1
    logsindex = list(range(len(loglist)))
    while (k <= n - 2):
        min_index = np.argmin(y)

        # update Z
        temp = []
        temp.extend(index_list[min_index])
        temp.append(y[min_index])
        cluster_size[n - 1 + k] = cluster_size[temp[0]] + cluster_size[temp[1]]
        temp.append(cluster_size[n - 1 + k])

        Z.append(temp)

        # get index of min in y
        item = index_list[min_index][::]
        record1 = []
        record2 = []
        for ele in index_list:
            if item[0] in ele:
                record1.append(index_list.index(ele))
                inde = ele.index(item[0])
                ele[inde] = n - 1 + k
            if item[1] in ele:  # here if/elif both works
                record2.append(index_list.index(ele))
                inde = ele.index(item[1])
                ele[inde] = n - 1 + k
            ele.sort()

        record = list(set(record1).union(set(record2)))

        merged1 = merge_log.update_merge([loglist[item[0]], loglist[item[1]]])
        # here the logsindex is changing
        diff = list(set(logsindex).difference(set(item)))  # diff is the node number need to be updated

        update_dist = dict()
        for ele in diff:
            (dist_act, dist_dfg) = dfg_dist.dfg_dist_calc(merged1, loglist[ele])
            tempdist = dist_act * alpha + dist_dfg * (1 - alpha)
            # tempdist = leven_dist_calc.leven_dist_avg(merged1, loglist[ele], percent, percent)
            update_dist[ele] = tempdist

        loglist.append(merged1)
        diff.append(n - 1 + k)
        logsindex = diff

        del (record1[record1.index(min_index)])
        del (record2[record2.index(min_index)])

        # for i in range(len(record1)):
        #     y[record1[i]] = (y[record1[i]]*cluster_size[item[0]] + y[record2[i]]*cluster_size[item[1]]) / (cluster_size[item[0]]+cluster_size[item[1]])
        for ele in record1:
            uindex = index_list[ele][0]  # record1 is the location if nodes in diff in the index_list
            y[ele] = update_dist[uindex]

        diff1 = list(set(range(len(index_list))).difference(set(record)))
        newindex = record1 + diff1
        newindex.sort()

        range_newindex = range(len(newindex))
        tempy = list(range_newindex)
        templist = list(range_newindex)
        for i in range_newindex:
            tempy[i] = y[newindex[i]]
            templist[i] = index_list[newindex[i]]

        index_list = templist
        y = tempy
        k = k + 1

    temp = []
    temp.extend(index_list[0])
    temp.append(y[0])

    cluster_size[n - 1 + k] = cluster_size[temp[0]] + cluster_size[temp[1]]
    temp.append(cluster_size[n - 1 + k])
    Z.append(temp)
    Z = np.array(Z)

    return Z


def linkage_avg(loglist, dist_mat, alpha, percent):
    index_list = []
    cluster_size = []
    for i in range(len(dist_mat)):
        cluster_size.append(len(loglist[i]))
        for j in range(i + 1, len(dist_mat)):
            index_list.append([i, j])

    y = squareform(dist_mat)
    n = len(dist_mat)  # The number of observations.
    Z = []
    cluster_size = dict(zip(range(n), cluster_size))  # record merged cluster size every step
    k = 1
    while (k <= n - 2):
        min_index = np.argmin(y)

        # update Z
        temp = []
        temp.extend(index_list[min_index])
        temp.append(y[min_index])
        cluster_size[n - 1 + k] = cluster_size[temp[0]] + cluster_size[temp[1]]
        temp.append(cluster_size[n - 1 + k])

        Z.append(temp)

        # get index of min in y
        item = index_list[min_index][::]
        record1 = []
        record2 = []
        for ele in index_list:
            if item[0] in ele:
                record1.append(index_list.index(ele))
                inde = ele.index(item[0])
                ele[inde] = n - 1 + k
            if item[1] in ele:  # here if/elif both works
                record2.append(index_list.index(ele))
                inde = ele.index(item[1])
                ele[inde] = n - 1 + k
            ele.sort()

        record = list(set(record1).union(set(record2)))

        del (record1[record1.index(min_index)])
        del (record2[record2.index(min_index)])

        for i in range(len(record1)):
            y[record1[i]] = (y[record1[i]] * cluster_size[item[0]] + y[record2[i]] * cluster_size[item[1]]) / (
                        cluster_size[item[0]] + cluster_size[item[1]])
        # for ele in record1:
        #     uindex = index_list[ele][0]  # record1 is the location if nodes in diff in the index_list
        #     y[ele] = update_dist[uindex]

        diff1 = list(set(range(len(index_list))).difference(set(record)))
        newindex = record1 + diff1
        newindex.sort()

        range_newindex = range(len(newindex))
        tempy = list(range_newindex)
        templist = list(range_newindex)
        for i in range_newindex:
            tempy[i] = y[newindex[i]]
            templist[i] = index_list[newindex[i]]

        index_list = templist
        y = tempy
        k = k + 1

    temp = []
    temp.extend(index_list[0])
    temp.append(y[0])

    cluster_size[n - 1 + k] = cluster_size[temp[0]] + cluster_size[temp[1]]
    temp.append(cluster_size[n - 1 + k])
    Z.append(temp)
    Z = np.array(Z)

    return Z


def linkage_DMM_update(loglist, dist_mat, alpha, percent):
    index_list = []
    for i in range(len(dist_mat)):
        for j in range(i + 1, len(dist_mat)):
            index_list.append([i, j])

    y = squareform(dist_mat)
    n = len(dist_mat)  # The number of observations.
    Z = []
    cluster_size = dict(zip(range(n), np.ones(n)))  # record merged cluster size every step
    k = 1
    logsindex = list(range(len(loglist)))
    while (k <= n - 2):
        min_index = np.argmin(y)

        # update Z
        temp = []
        temp.extend(index_list[min_index])
        temp.append(y[min_index])
        cluster_size[n - 1 + k] = cluster_size[temp[0]] + cluster_size[temp[1]]
        temp.append(cluster_size[n - 1 + k])

        Z.append(temp)

        # get index of min in y
        item = index_list[min_index][::]
        record1 = []
        record2 = []
        for ele in index_list:
            if item[0] in ele:
                record1.append(index_list.index(ele))
                inde = ele.index(item[0])
                ele[inde] = n - 1 + k
            if item[1] in ele:  # here if/elif both works
                record2.append(index_list.index(ele))
                inde = ele.index(item[1])
                ele[inde] = n - 1 + k
            ele.sort()

        record = list(set(record1).union(set(record2)))

        merged1 = merge_log.update_merge([loglist[item[0]], loglist[item[1]]])
        # here the logsindex is changing
        diff = list(set(logsindex).difference(set(item)))  # diff is the node number need to be updated

        update_dist = dict()
        for ele in diff:
            dist_act = act_dist_calc.act_sim_percent(merged1, loglist[ele], percent, percent)
            dist_suc = suc_dist_calc.suc_sim_percent(merged1, loglist[ele], percent, percent)
            tempdist = dist_act * alpha + dist_suc * (1 - alpha)
            # tempdist = leven_dist_calc.leven_dist_avg(merged1, loglist[ele], percent, percent)
            update_dist[ele] = tempdist

        loglist.append(merged1)
        diff.append(n - 1 + k)
        logsindex = diff

        del (record1[record1.index(min_index)])
        del (record2[record2.index(min_index)])

        # for i in range(len(record1)):
        #     y[record1[i]] = (y[record1[i]]*cluster_size[item[0]] + y[record2[i]]*cluster_size[item[1]]) / (cluster_size[item[0]]+cluster_size[item[1]])
        for ele in record1:
            uindex = index_list[ele][0]  # record1 is the location if nodes in diff in the index_list
            y[ele] = update_dist[uindex]

        diff1 = list(set(range(len(index_list))).difference(set(record)))
        newindex = record1 + diff1
        newindex.sort()

        range_newindex = range(len(newindex))
        tempy = list(range_newindex)
        templist = list(range_newindex)
        for i in range_newindex:
            tempy[i] = y[newindex[i]]
            templist[i] = index_list[newindex[i]]

        index_list = templist
        y = tempy
        k = k + 1

    temp = []
    temp.extend(index_list[0])
    temp.append(y[0])

    cluster_size[n - 1 + k] = cluster_size[temp[0]] + cluster_size[temp[1]]
    temp.append(cluster_size[n - 1 + k])
    Z.append(temp)
    Z = np.array(Z)

    return Z


def linkage_DMM_update_leven(loglist, dist_mat, alpha, percent):
    index_list = []
    for i in range(len(dist_mat)):
        for j in range(i + 1, len(dist_mat)):
            index_list.append([i, j])

    y = squareform(dist_mat)
    n = len(dist_mat)  # The number of observations.
    Z = []
    cluster_size = dict(zip(range(n), np.ones(n)))  # record merged cluster size every step
    k = 1
    logsindex = list(range(len(loglist)))
    while (k <= n - 2):
        min_index = np.argmin(y)

        # update Z
        temp = []
        temp.extend(index_list[min_index])
        temp.append(y[min_index])
        cluster_size[n - 1 + k] = cluster_size[temp[0]] + cluster_size[temp[1]]
        temp.append(cluster_size[n - 1 + k])

        Z.append(temp)

        # get index of min in y
        item = index_list[min_index][::]
        record1 = []
        record2 = []
        for ele in index_list:
            if item[0] in ele:
                record1.append(index_list.index(ele))
                inde = ele.index(item[0])
                ele[inde] = n - 1 + k
            if item[1] in ele:  # here if/elif both works
                record2.append(index_list.index(ele))
                inde = ele.index(item[1])
                ele[inde] = n - 1 + k
            ele.sort()

        record = list(set(record1).union(set(record2)))

        merged1 = merge_log.update_merge([loglist[item[0]], loglist[item[1]]])
        # here the logsindex is changing
        diff = list(set(logsindex).difference(set(item)))  # diff is the node number need to be updated

        update_dist = dict()
        for ele in diff:
            tempdist = leven_dist_calc.leven_dist(merged1, loglist[ele], percent, percent)
            # tempdist = leven_dist_calc.leven_dist_avg(merged1, loglist[ele], percent, percent)
            update_dist[ele] = tempdist

        loglist.append(merged1)
        diff.append(n - 1 + k)
        logsindex = diff

        del (record1[record1.index(min_index)])
        del (record2[record2.index(min_index)])

        # for i in range(len(record1)):
        #     y[record1[i]] = (y[record1[i]]*cluster_size[item[0]] + y[record2[i]]*cluster_size[item[1]]) / (cluster_size[item[0]]+cluster_size[item[1]])
        for ele in record1:
            uindex = index_list[ele][0]  # record1 is the location if nodes in diff in the index_list
            y[ele] = update_dist[uindex]

        diff1 = list(set(range(len(index_list))).difference(set(record)))
        newindex = record1 + diff1
        newindex.sort()

        range_newindex = range(len(newindex))
        tempy = list(range_newindex)
        templist = list(range_newindex)
        for i in range_newindex:
            tempy[i] = y[newindex[i]]
            templist[i] = index_list[newindex[i]]

        index_list = templist
        y = tempy
        k = k + 1

    temp = []
    temp.extend(index_list[0])
    temp.append(y[0])

    cluster_size[n - 1 + k] = cluster_size[temp[0]] + cluster_size[temp[1]]
    temp.append(cluster_size[n - 1 + k])
    Z.append(temp)
    Z = np.array(Z)

    return Z



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.algo.discovery.causal import algorithm as causal_algorithm
from pm4py.algo.discovery.causal.algorithm import CAUSAL_ALPHA


class ClassicAlphaAbstraction:
    """
    Class representing the basic abstraction of the alpha miner.
    The class covers start- and end attributes, the directly follows relation, the parallel relation and the causal
    relation.
    """

    def __init__(self, start_activities, end_activities, dfg, activity_key="concept:name"):
        self.__activity_key = activity_key
        self.__start_activities = start_activities
        self.__end_activities = end_activities
        self.__dfg = dfg
        self.__causal_relations = {k: v for k, v in causal_algorithm.apply(self.dfg, variant=CAUSAL_ALPHA).items() if
                                   v > 0}.keys()
        self.__parallel = {(f, t) for (f, t) in self.dfg if (t, f) in self.dfg}

    def __get_causal_relation(self):
        return self.__causal_relations

    def __get_start_activities(self):
        return self.__start_activities

    def __get_end_activities(self):
        return self.__end_activities

    def __get_directly_follows_graph(self):
        return self.__dfg

    def __get_parallel_relation(self):
        return self.__parallel

    def __get_activity_key(self):
        return self.__activity_key

    start_activities = property(__get_start_activities)
    end_activities = property(__get_end_activities)
    dfg = property(__get_directly_follows_graph)
    causal_relation = property(__get_causal_relation)
    parallel_relation = property(__get_parallel_relation)
    activity_key = property(__get_activity_key)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import uuid
from enum import Enum
from collections import Counter
from pm4py.util import nx_utils


DEFAULT_PROCESS = str(uuid.uuid4())


class Marking(Counter):
    pass

    # required = Counter()

    def __hash__(self):
        r = 0
        for p in self.items():
            r += 31 * hash(p[0]) * p[1]
        return r

    def __eq__(self, other):
        if not self.keys() == other.keys():
            return False
        for p in self.keys():
            if other.get(p) != self.get(p):
                return False
        return True

    def __le__(self, other):
        if not self.keys() <= other.keys():
            return False
        for p in self.keys():
            if sum(other.get(p)) < sum(self.get(p)):
                return False
        return True

    def __add__(self, other):
        m = Marking()
        for p in self.items():
            m[p[0]] = p[1]
        for p in other.items():
            m[p[0]] += p[1]
        return m

    def __sub__(self, other):
        m = Marking()
        for p in self.items():
            m[p[0]] = p[1]
        for p in other.items():
            m[p[0]] -= p[1]
            if m[p[0]] == 0:
                del m[p[0]]
        return m

    def __repr__(self):
        # return str([str(p.name) + ":" + str(self.get(p)) for p in self.keys()])
        # The previous representation had a bug, it took into account the order
        # of the places with tokens
        return str(
            [
                str(p.id) + ":" + str(self.get(p))
                for p in sorted(list(self.keys()), key=lambda x: x.id)
            ]
        )

    def __deepcopy__(self, memodict={}):
        marking = Marking()
        memodict[id(self)] = marking
        for node in self:
            node_occ = self[node]
            new_node = (
                memodict[id(node)]
                if id(node) in memodict
                else BPMN.BPMNNode(node.id, node.name)
            )
            marking[new_node] = node_occ
        return marking


class BPMNNodeLayout(object):
    def __init__(self):
        self.__x = 0
        self.__y = 0
        self.__width = 100
        self.__height = 100

    def set_x(self, x):
        self.__x = x

    def set_y(self, y):
        self.__y = y

    def get_x(self):
        return self.__x

    def get_y(self):
        return self.__y

    def get_width(self):
        return self.__width

    def set_width(self, width):
        self.__width = width

    def get_height(self):
        return self.__height

    def set_height(self, height):
        self.__height = height


class BPMNEdgeLayout(object):
    def __init__(self):
        self.__waypoints = [(0, 0), (0, 0)]

    def add_waypoint(self, waypoint):
        self.__waypoints.append(waypoint)

    def del_waypoints(self):
        self.__waypoints = list()

    def get_waypoints(self):
        return self.__waypoints


class BPMNLayout(object):
    def __init__(self):
        self.layout_dict = {}

    def get(self, n):
        if n not in self.layout_dict:
            if isinstance(n, BPMN.BPMNNode):
                self.layout_dict[n] = BPMNNodeLayout()
            elif isinstance(n, BPMN.Flow):
                self.layout_dict[n] = BPMNEdgeLayout()
        return self.layout_dict[n]


class BPMN(object):
    class BPMNNode(object):
        def __init__(
            self, id="", name="", in_arcs=None, out_arcs=None, process=None
        ):
            self.__id = ("id" + str(uuid.uuid4())) if id == "" else id
            self.__name = name
            self.__in_arcs = list() if in_arcs is None else in_arcs
            self.__out_arcs = list() if out_arcs is None else out_arcs
            self.__process = DEFAULT_PROCESS if process is None else process
            self.__layout = BPMNLayout()

        def get_id(self):
            return self.__id

        def get_name(self):
            return self.__name

        def set_name(self, name):
            self.__name = name

        def get_in_arcs(self):
            return self.__in_arcs

        def get_out_arcs(self):
            return self.__out_arcs

        def add_in_arc(self, in_arc):
            if in_arc not in self.__in_arcs:
                self.__in_arcs.append(in_arc)

        def add_out_arc(self, out_arc):
            if out_arc not in self.__out_arcs:
                self.__out_arcs.append(out_arc)

        def remove_in_arc(self, in_arc):
            self.__in_arcs.remove(in_arc)

        def remove_out_arc(self, out_arc):
            self.__out_arcs.remove(out_arc)

        def get_process(self):
            return self.__process

        def set_process(self, process):
            self.__process = process

        def set_x(self, x):
            return self.__layout.get(self).set_x(x)

        def set_y(self, y):
            return self.__layout.get(self).set_y(y)

        def get_x(self):
            return self.__layout.get(self).get_x()

        def get_y(self):
            return self.__layout.get(self).get_y()

        def get_width(self):
            return self.__layout.get(self).get_width()

        def set_width(self, width):
            return self.__layout.get(self).set_width(width)

        def get_height(self):
            return self.__layout.get(self).get_height()

        def set_height(self, height):
            return self.__layout.get(self).set_height(height)

        def get_layout(self):
            return self.__layout

        def set_layout(self, layout):
            self.__layout = layout

        def __hash__(self):
            return hash(self.id)

        def __eq__(self, other):
            # keep the ID for now in places
            return hash(self) == hash(other)

        def __repr__(self):
            return str(self.__id + "@" + self.__name)

        def __str__(self):
            return self.__repr__()

        name = property(get_name)
        id = property(get_id)
        in_arcs = property(get_in_arcs)
        out_arcs = property(get_out_arcs)
        process = property(get_process, set_process)

    class Collaboration(BPMNNode):
        def __init__(
            self, id="", name="", in_arcs=None, out_arcs=None, process=None
        ):
            BPMN.BPMNNode.__init__(
                self, id, name, in_arcs, out_arcs, process=process
            )

    class Participant(BPMNNode):
        def __init__(
            self,
            id="",
            name="",
            in_arcs=None,
            out_arcs=None,
            process=None,
            process_ref=None,
        ):
            self.process_ref = process_ref
            BPMN.BPMNNode.__init__(
                self, id, name, in_arcs, out_arcs, process=process
            )

    class TextAnnotation(BPMNNode):
        def __init__(
            self,
            id="",
            name="",
            in_arcs=None,
            out_arcs=None,
            process=None,
            text=None,
        ):
            self.text = text
            BPMN.BPMNNode.__init__(
                self, id, name, in_arcs, out_arcs, process=process
            )

    class Event(BPMNNode):
        def __init__(
            self, id="", name="", in_arcs=None, out_arcs=None, process=None
        ):
            BPMN.BPMNNode.__init__(
                self, id, name, in_arcs, out_arcs, process=process
            )

    class StartEvent(Event):
        def __init__(
            self,
            id="",
            isInterrupting=False,
            name="",
            parallelMultiple=False,
            in_arcs=None,
            out_arcs=None,
            process=None,
        ):
            BPMN.Event.__init__(
                self, id, name, in_arcs, out_arcs, process=process
            )
            self.__isInterrupting = isInterrupting
            self.__parallelMultiple = parallelMultiple

        def get_isInterrupting(self):
            return self.__isInterrupting

        def get_parallelMultiple(self):
            return self.__parallelMultiple

    class NormalStartEvent(StartEvent):
        def __init__(
            self,
            id="",
            isInterrupting=False,
            name="",
            parallelMultiple=False,
            in_arcs=None,
            out_arcs=None,
            process=None,
        ):
            BPMN.StartEvent.__init__(
                self,
                id,
                isInterrupting,
                name,
                parallelMultiple,
                in_arcs,
                out_arcs,
                process=process,
            )

    class MessageStartEvent(StartEvent):
        def __init__(
            self,
            id="",
            isInterrupting=False,
            name="",
            parallelMultiple=False,
            in_arcs=None,
            out_arcs=None,
            process=None,
        ):
            BPMN.StartEvent.__init__(
                self,
                id,
                isInterrupting,
                name,
                parallelMultiple,
                in_arcs,
                out_arcs,
                process=process,
            )

    class IntermediateCatchEvent(Event):
        def __init__(
            self, id="", name="", in_arcs=None, out_arcs=None, process=None
        ):
            BPMN.Event.__init__(
                self, id, name, in_arcs, out_arcs, process=process
            )

    class MessageIntermediateCatchEvent(IntermediateCatchEvent):
        def __init__(
            self, id="", name="", in_arcs=None, out_arcs=None, process=None
        ):
            BPMN.IntermediateCatchEvent.__init__(
                self, id, name, in_arcs, out_arcs, process=process
            )

    class ErrorIntermediateCatchEvent(IntermediateCatchEvent):
        def __init__(
            self, id="", name="", in_arcs=None, out_arcs=None, process=None
        ):
            BPMN.IntermediateCatchEvent.__init__(
                self, id, name, in_arcs, out_arcs, process=process
            )

    class CancelIntermediateCatchEvent(IntermediateCatchEvent):
        def __init__(
            self, id="", name="", in_arcs=None, out_arcs=None, process=None
        ):
            BPMN.IntermediateCatchEvent.__init__(
                self, id, name, in_arcs, out_arcs, process=process
            )

    class BoundaryEvent(Event):
        def __init__(
            self,
            id="",
            name="",
            in_arcs=None,
            out_arcs=None,
            process=None,
            activity=None,
        ):
            self.__activity = activity
            BPMN.Event.__init__(
                self, id, name, in_arcs, out_arcs, process=process
            )

        def get_activity(self):
            return self.__activity

    class MessageBoundaryEvent(BoundaryEvent):
        def __init__(
            self,
            id="",
            name="",
            in_arcs=None,
            out_arcs=None,
            process=None,
            activity=None,
        ):
            BPMN.BoundaryEvent.__init__(
                self,
                id,
                name,
                in_arcs,
                out_arcs,
                process=process,
                activity=activity,
            )

    class ErrorBoundaryEvent(BoundaryEvent):
        def __init__(
            self,
            id="",
            name="",
            in_arcs=None,
            out_arcs=None,
            process=None,
            activity=None,
        ):
            BPMN.BoundaryEvent.__init__(
                self,
                id,
                name,
                in_arcs,
                out_arcs,
                process=process,
                activity=activity,
            )

    class CancelBoundaryEvent(BoundaryEvent):
        def __init__(
            self,
            id="",
            name="",
            in_arcs=None,
            out_arcs=None,
            process=None,
            activity=None,
        ):
            BPMN.BoundaryEvent.__init__(
                self,
                id,
                name,
                in_arcs,
                out_arcs,
                process=process,
                activity=activity,
            )

    class IntermediateThrowEvent(Event):
        def __init__(
            self, id="", name="", in_arcs=None, out_arcs=None, process=None
        ):
            BPMN.Event.__init__(
                self, id, name, in_arcs, out_arcs, process=process
            )

    class MessageIntermediateThrowEvent(IntermediateThrowEvent):
        def __init__(
            self, id="", name="", in_arcs=None, out_arcs=None, process=None
        ):
            BPMN.IntermediateThrowEvent.__init__(
                self, id, name, in_arcs, out_arcs, process=process
            )

    class NormalIntermediateThrowEvent(IntermediateThrowEvent):
        def __init__(
            self, id="", name="", in_arcs=None, out_arcs=None, process=None
        ):
            BPMN.IntermediateThrowEvent.__init__(
                self, id, name, in_arcs, out_arcs, process=process
            )

    class EndEvent(Event):
        def __init__(
            self, id="", name="", in_arcs=None, out_arcs=None, process=None
        ):
            BPMN.Event.__init__(
                self, id, name, in_arcs, out_arcs, process=process
            )

    class NormalEndEvent(EndEvent):
        def __init__(
            self, id="", name="", in_arcs=None, out_arcs=None, process=None
        ):
            BPMN.EndEvent.__init__(
                self, id, name, in_arcs, out_arcs, process=process
            )

    class MessageEndEvent(EndEvent):
        def __init__(
            self, id="", name="", in_arcs=None, out_arcs=None, process=None
        ):
            BPMN.EndEvent.__init__(
                self, id, name, in_arcs, out_arcs, process=process
            )

    class TerminateEndEvent(EndEvent):
        def __init__(
            self, id="", name="", in_arcs=None, out_arcs=None, process=None
        ):
            BPMN.EndEvent.__init__(
                self, id, name, in_arcs, out_arcs, process=process
            )

    class ErrorEndEvent(EndEvent):
        def __init__(
            self, id="", name="", in_arcs=None, out_arcs=None, process=None
        ):
            BPMN.EndEvent.__init__(
                self, id, name, in_arcs, out_arcs, process=process
            )

    class CancelEndEvent(EndEvent):
        def __init__(
            self, id="", name="", in_arcs=None, out_arcs=None, process=None
        ):
            BPMN.EndEvent.__init__(
                self, id, name, in_arcs, out_arcs, process=process
            )

    class Activity(BPMNNode):
        def __init__(
            self, id="", name="", in_arcs=None, out_arcs=None, process=None
        ):
            BPMN.BPMNNode.__init__(
                self, id, name, in_arcs, out_arcs, process=process
            )

    class Task(Activity):
        def __init__(
            self, id="", name="", in_arcs=None, out_arcs=None, process=None
        ):
            BPMN.Activity.__init__(
                self, id, name, in_arcs, out_arcs, process=process
            )

    class UserTask(Task):
        def __init__(
            self, id="", name="", in_arcs=None, out_arcs=None, process=None
        ):
            BPMN.Task.__init__(
                self, id, name, in_arcs, out_arcs, process=process
            )

    class SendTask(Task):
        def __init__(
            self, id="", name="", in_arcs=None, out_arcs=None, process=None
        ):
            BPMN.Task.__init__(
                self, id, name, in_arcs, out_arcs, process=process
            )

    class SubProcess(Activity):
        def __init__(
            self,
            id="",
            name="",
            in_arcs=None,
            out_arcs=None,
            process=None,
            depth=None,
        ):
            self.__depth = depth
            BPMN.Activity.__init__(
                self, id, name, in_arcs, out_arcs, process=process
            )

        def get_depth(self):
            return self.__depth

    class Gateway(BPMNNode):
        class Direction(Enum):
            UNSPECIFIED = "Unspecified"
            DIVERGING = "Diverging"
            CONVERGING = "Converging"

        def __init__(
            self,
            id="",
            name="",
            gateway_direction=Direction.UNSPECIFIED,
            in_arcs=None,
            out_arcs=None,
            process=None,
        ):
            BPMN.BPMNNode.__init__(
                self, id, name, in_arcs, out_arcs, process=process
            )
            self.__gateway_direction = gateway_direction

        def get_gateway_direction(self):
            return self.__gateway_direction

        def set_gateway_direction(self, direction):
            self.__gateway_direction = direction

    class ParallelGateway(Gateway):
        def __init__(
            self,
            id="",
            name="",
            gateway_direction=None,
            in_arcs=None,
            out_arcs=None,
            process=None,
        ):
            gateway_direction = (
                gateway_direction
                if gateway_direction is not None
                else BPMN.Gateway.Direction.UNSPECIFIED
            )
            BPMN.Gateway.__init__(
                self,
                id,
                name,
                gateway_direction,
                in_arcs,
                out_arcs,
                process=process,
            )

    class ExclusiveGateway(Gateway):
        def __init__(
            self,
            id="",
            name="",
            gateway_direction=None,
            in_arcs=None,
            out_arcs=None,
            process=None,
        ):
            gateway_direction = (
                gateway_direction
                if gateway_direction is not None
                else BPMN.Gateway.Direction.UNSPECIFIED
            )
            BPMN.Gateway.__init__(
                self,
                id,
                name,
                gateway_direction,
                in_arcs,
                out_arcs,
                process=process,
            )

    class InclusiveGateway(Gateway):
        def __init__(
            self,
            id="",
            name="",
            gateway_direction=None,
            in_arcs=None,
            out_arcs=None,
            process=None,
        ):
            gateway_direction = (
                gateway_direction
                if gateway_direction is not None
                else BPMN.Gateway.Direction.UNSPECIFIED
            )
            BPMN.Gateway.__init__(
                self,
                id,
                name,
                gateway_direction,
                in_arcs,
                out_arcs,
                process=process,
            )

    class EventBasedGateway(Gateway):
        def __init__(
            self,
            id="",
            name="",
            gateway_direction=None,
            in_arcs=None,
            out_arcs=None,
            process=None,
        ):
            gateway_direction = (
                gateway_direction
                if gateway_direction is not None
                else BPMN.Gateway.Direction.UNSPECIFIED
            )
            BPMN.Gateway.__init__(
                self,
                id,
                name,
                gateway_direction,
                in_arcs,
                out_arcs,
                process=process,
            )

    class Flow(object):
        def __init__(self, source, target, id="", name="", process=None):
            self.__id = uuid.uuid4() if id == "" else id
            self.__name = name
            self.__source = source
            source.add_out_arc(self)
            self.__target = target
            target.add_in_arc(self)
            self.__process = DEFAULT_PROCESS if process is None else process
            self.__layout = BPMNLayout()

        def get_id(self):
            return self.__id

        def get_name(self):
            return self.__name

        def get_source(self):
            return self.__source

        def get_target(self):
            return self.__target

        def get_process(self):
            return self.__process

        def set_process(self, process):
            self.__process = process

        def add_waypoint(self, waypoint):
            return self.__layout.get(self).add_waypoint(waypoint)

        def del_waypoints(self):
            return self.__layout.get(self).del_waypoints()

        def get_waypoints(self):
            return self.__layout.get(self).get_waypoints()

        def get_layout(self):
            return self.__layout

        def set_layout(self, layout):
            self.__layout = layout

        def __repr__(self):
            u_id = (
                str(self.__source.get_id())
                + "@"
                + str(self.__source.get_name())
            )
            v_id = (
                str(self.__target.get_id())
                + "@"
                + str(self.__target.get_name())
            )
            return u_id + " -> " + v_id

        def __str__(self):
            return self.__repr__()

        source = property(get_source)
        target = property(get_target)

    class SequenceFlow(Flow):
        def __init__(self, source, target, id="", name="", process=None):
            BPMN.Flow.__init__(
                self, source, target, id=id, name=name, process=process
            )

    class MessageFlow(Flow):
        def __init__(self, source, target, id="", name="", process=None):
            BPMN.Flow.__init__(
                self, source, target, id=id, name=name, process=process
            )

    class Association(Flow):
        def __init__(self, source, target, id="", name="", process=None):
            BPMN.Flow.__init__(
                self, source, target, id=id, name=name, process=process
            )

    def __init__(self, process_id=None, name="", nodes=None, flows=None):
        self.__process_id = (
            str(uuid.uuid4()) if process_id is None else process_id
        )

        self.__name = name
        self.__graph = nx_utils.MultiDiGraph()
        self.__nodes = set() if nodes is None else nodes
        self.__flows = set() if flows is None else flows
        self.__layout = BPMNLayout()

        if nodes is not None:
            for node in nodes:
                node.set_layout(self.get_layout())
                self.__graph.add_node(node)
        if flows is not None:
            for flow in flows:
                flow.set_layout(self.get_layout())
                self.__graph.add_edge(flow.get_source(), flow.get_target())

    def get_process_id(self):
        return self.__process_id

    def set_process_id(self, process_id):
        self.__process_id = process_id

    def get_nodes(self):
        return self.__nodes

    def get_flows(self):
        return self.__flows

    def get_graph(self):
        return self.__graph

    def get_name(self):
        return self.__name

    def set_name(self, name: str):
        self.__name = name

    def add_node(self, node):
        node.set_layout(self.get_layout())
        self.__nodes.add(node)
        self.__graph.add_node(node)

    def remove_node(self, node):
        if node in self.__nodes:
            self.__nodes.remove(node)
            self.__graph.remove_node(node)

    def remove_flow(self, flow):
        source = flow.get_source()
        target = flow.get_target()
        if source in self.__nodes:
            source.remove_out_arc(flow)
        if target in self.__nodes:
            target.remove_in_arc(flow)
        self.__flows.remove(flow)
        self.__graph.remove_edge(source, target)

    def add_flow(self, flow):
        if not isinstance(flow, BPMN.Flow):
            raise Exception()
        flow.set_layout(self.get_layout())
        source = flow.get_source()
        target = flow.get_target()
        if source not in self.__nodes:
            self.add_node(source)
        if target not in self.__nodes:
            self.add_node(target)
        self.__flows.add(flow)
        self.__graph.add_edge(
            source, target, id=flow.get_id(), name=flow.get_name()
        )
        source.add_out_arc(flow)
        target.add_in_arc(flow)

    def get_layout(self):
        return self.__layout

    def set_layout(self, layout):
        self.__layout = layout
        for n in self.__nodes:
            n.set_layout(layout)
        for e in self.__flows:
            e.set_layout(layout)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.util import exec_utils
from enum import Enum
from pm4py.util import constants, xes_constants, pandas_utils
from pm4py.objects.conversion.log import converter
from pm4py.algo.discovery.correlation_mining import util as cm_util
import numpy as np
from collections import Counter
from typing import Optional, Dict, Any, Union, Tuple
from pm4py.objects.log.obj import EventLog, EventStream
import pandas as pd


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY
    START_TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_START_TIMESTAMP_KEY
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY
    INDEX_KEY = "index_key"


DEFAULT_INDEX_KEY = "@@@index"


def apply(log: Union[EventLog, EventStream, pd.DataFrame], parameters: Optional[Dict[Union[str, Parameters], Any]] = None) -> Tuple[Dict[Tuple[str, str], int], Dict[Tuple[str, str], float]]:
    """
    Novel approach of correlation mining, that creates the PS-matrix and the duration matrix
    using the order list of events of each trace of the log

    Parameters
    -------------
    log
        Event log
    parameters
        Parameters

    Returns
    ---------------
    dfg
        DFG
    performance_dfg
        Performance DFG (containing the estimated performance for the arcs)
    """
    traces_list, trace_grouped_list, activities, activities_counter = preprocess_log(log, activities=None,
                                                                                     activities_counter=None)

    PS_matrix, duration_matrix = get_PS_duration_matrix(activities, trace_grouped_list, parameters=parameters)

    return resolve_lp_get_dfg(PS_matrix, duration_matrix, activities, activities_counter)


def resolve_lp_get_dfg(PS_matrix, duration_matrix, activities, activities_counter):
    """
    Resolves a LP problem to get a DFG

    Parameters
    ---------------
    PS_matrix
        Precede-succeed matrix
    duration_matrix
        Duration matrix
    activities
        List of activities of the log
    activities_counter
        Counter for the activities of the log

    Returns
    ---------------
    dfg
        Frequency DFG
    performance_dfg
        Performance DFG
    """
    C_matrix = cm_util.get_c_matrix(PS_matrix, duration_matrix, activities, activities_counter)
    dfg, performance_dfg = cm_util.resolve_LP(C_matrix, duration_matrix, activities, activities_counter)
    return dfg, performance_dfg


def get_PS_duration_matrix(activities, trace_grouped_list, parameters=None):
    """
    Gets the precede-succeed matrix

    Parameters
    --------------
    activities
        Activities
    trace_grouped_list
        Grouped list of simplified traces (per activity)
    parameters
        Parameters of the algorithm

    Returns
    --------------
    PS_matrix
        precede-succeed matrix
    duration_matrix
        Duration matrix
    """
    if parameters is None:
        parameters = {}

    timestamp_key = exec_utils.get_param_value(Parameters.TIMESTAMP_KEY, parameters,
                                               xes_constants.DEFAULT_TIMESTAMP_KEY)
    start_timestamp_key = exec_utils.get_param_value(Parameters.START_TIMESTAMP_KEY, parameters,
                                                     xes_constants.DEFAULT_TIMESTAMP_KEY)

    PS_matrix = get_precede_succeed_matrix(activities, trace_grouped_list, timestamp_key, start_timestamp_key)
    duration_matrix = get_duration_matrix(activities, trace_grouped_list, timestamp_key, start_timestamp_key)

    return PS_matrix, duration_matrix


def preprocess_log(log, activities=None, activities_counter=None, parameters=None):
    """
    Preprocess the log to get a grouped list of simplified traces (per activity)

    Parameters
    --------------
    log
        Log object
    activities
        (if provided) activities of the log
    activities_counter
        (if provided) counter of the activities of the log
    parameters
        Parameters of the algorithm

    Returns
    --------------
    traces_list
        List of simplified traces of the log
    trace_grouped_list
        Grouped list of simplified traces (per activity)
    activities
        Activities of the log
    activities_counter
        Activities counter
    """
    if parameters is None:
        parameters = {}

    activity_key = exec_utils.get_param_value(Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY)
    timestamp_key = exec_utils.get_param_value(Parameters.TIMESTAMP_KEY, parameters,
                                               xes_constants.DEFAULT_TIMESTAMP_KEY)
    start_timestamp_key = exec_utils.get_param_value(Parameters.START_TIMESTAMP_KEY, parameters,
                                                     xes_constants.DEFAULT_TIMESTAMP_KEY)
    caseid_key = exec_utils.get_param_value(Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME)
    index_key = exec_utils.get_param_value(Parameters.INDEX_KEY, parameters, DEFAULT_INDEX_KEY)

    if pandas_utils.check_is_pandas_dataframe(log):
        # keep only the two columns before conversion
        log = log[list(set([activity_key, timestamp_key, start_timestamp_key, caseid_key]))]

    log = converter.apply(log, variant=converter.Variants.TO_EVENT_LOG, parameters=parameters)

    traces_list = []
    for trace in log:
        trace_stream = [
            {activity_key: trace[i][activity_key], timestamp_key: trace[i][timestamp_key].timestamp(),
             start_timestamp_key: trace[i][start_timestamp_key].timestamp(), index_key: i} for
            i in range(len(trace))]
        trace_stream = sorted(trace_stream, key=lambda x: (x[start_timestamp_key], x[timestamp_key], x[index_key]))
        traces_list.append(trace_stream)

    if activities is None:
        activities = sorted(list(set(y[activity_key] for x in traces_list for y in x)))

    trace_grouped_list = []
    for trace in traces_list:
        gr = []
        for act in activities:
            act_gr = [x for x in trace if x[activity_key] == act]
            gr.append(act_gr)
        trace_grouped_list.append(gr)

    if activities_counter is None:
        activities_counter = Counter(y[activity_key] for x in traces_list for y in x)

    return traces_list, trace_grouped_list, activities, activities_counter


def get_precede_succeed_matrix(activities, trace_grouped_list, timestamp_key, start_timestamp_key):
    """
    Calculates the precede succeed matrix

    Parameters
    ---------------
    activities
        Sorted list of activities of the log
    trace_grouped_list
        A list of lists of lists, containing for each trace and each activity the events having such activity
    timestamp_key
        The key to be used as timestamp
    start_timestamp_key
        The key to be used as start timestamp

    Returns
    ---------------
    mat
        The precede succeed matrix
    """
    ret = np.zeros((len(activities), len(activities)))
    for i in range(len(activities)):
        for j in range(len(activities)):
            if not i == j:
                count = 0
                total = 0
                for tr in trace_grouped_list:
                    ai = [x[timestamp_key] for x in tr[i]]
                    aj = [x[start_timestamp_key] for x in tr[j]]
                    if ai and aj:
                        total += len(ai) * len(aj)
                        k = 0
                        z = 0
                        while k < len(ai):
                            while z < len(aj):
                                if ai[k] < aj[z]:
                                    break
                                z = z + 1
                            count = count + (len(aj) - z)
                            k = k + 1
                if total > 0:
                    ret[i, j] = count / float(total)
    return ret


def get_duration_matrix(activities, trace_grouped_list, timestamp_key, start_timestamp_key):
    """
    Calculates the duration matrix

    Parameters
    --------------
    activities
        Sorted list of activities of the log
    trace_grouped_list
        A list of lists of lists, containing for each trace and each activity the events having such activity
    timestamp_key
        The key to be used as timestamp
    start_timestamp_key
        The key to be used as start timestamp

    Returns
    --------------
    mat
        The duration matrix
    """
    from statistics import mean

    ret = np.zeros((len(activities), len(activities)))
    for i in range(len(activities)):
        for j in range(len(activities)):
            if not i == j:
                tm0 = []
                tm1 = []
                for tr in trace_grouped_list:
                    ai = [x[timestamp_key] for x in tr[i]]
                    aj = [x[start_timestamp_key] for x in tr[j]]
                    if ai and aj:
                        tm0 = cm_util.calculate_time_match_fifo(ai, aj, times0=tm0)
                        tm1 = cm_util.calculate_time_match_rlifo(ai, aj, times1=tm1)
                td0 = mean([x[1] - x[0] for x in tm0]) if tm0 else 0
                td1 = mean([x[1] - x[0] for x in tm1]) if tm1 else 0
                ret[i, j] = min(td0, td1)
    return ret



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.visualization.transition_system.util import visualize_graphviz
from enum import Enum
from typing import Optional, Dict, Any, Union
from pm4py.objects.transition_system.obj import TransitionSystem
import graphviz


class Parameters(Enum):
    FORMAT = "format"
    SHOW_LABELS = "show_labels"
    SHOW_NAMES = "show_names"
    FORCE_NAMES = "force_names"
    FILLCOLORS = "fillcolors"
    FONT_SIZE = "font_size"
    ENABLE_GRAPH_TITLE = "enable_graph_title"
    GRAPH_TITLE = "graph_title"


def apply(tsys: TransitionSystem, parameters: Optional[Dict[Union[str, Parameters], Any]] = None) -> graphviz.Digraph:
    """
    Get visualization of a Transition System

    Parameters
    -----------
    tsys
        Transition system
    parameters
        Optional parameters of the algorithm

    Returns
    ----------
    gviz
        Graph visualization
    """

    gviz = visualize_graphviz.visualize(tsys, parameters=parameters)
    return gviz



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from collections import Counter
from typing import List, Dict, Any


from enum import Enum


class Outputs(Enum):
    DFG = "dfg"
    SEQUENCE = "sequence"
    PARALLEL = "parallel"
    START_ACTIVITIES = "start_activities"
    END_ACTIVITIES = "end_activities"
    ACTIVITIES = "activities"
    SKIPPABLE = "skippable"
    ACTIVITIES_ALWAYS_HAPPENING = "activities_always_happening"
    MIN_TRACE_LENGTH = "min_trace_length"
    TRACE = "trace"


DFG = "dfg"
FOOTPRINTS_KEY = "footprints"
START_ACTIVITIES = "start_activities"
END_ACTIVITIES = "end_activities"
SEQUENCE = "sequence"
PARALLEL = "parallel"
IS_FOOTPRINTS_FIT = "is_footprints_fit"


def fp_fitness(fp_log, fp_model, conf_results, parameters=None):
    """
    Calculates the footprints fitness provided the footprints of the log,
    and the result of footprints conformance (applied to the entire log)

    Parameters
    ---------------
    fp_log
        Footprints of the log
    fp_model
        Footprints of the model
    conf_results
        Footprints conformance (applied to the entire log)
    parameters
        Parameters of the algorithm

    Returns
    ---------------
    fitness
        Fitness value (between 0.0 and 1.0)
    """
    if parameters is None:
        parameters = {}

    fit_traces = None
    if isinstance(conf_results, list):
        fit_traces = len([x for x in conf_results if x[IS_FOOTPRINTS_FIT]])/len(conf_results) * 100.0

    fp_log = flatten_fp(fp_log)
    conf_results = flatten_conf(conf_results)

    dfg = fp_log[DFG]
    num_sequence_log = len(fp_log[SEQUENCE])
    num_parallel_log = len(fp_log[PARALLEL])
    num_start_activities_log = len(fp_log[START_ACTIVITIES])
    num_end_activities_log = len(fp_log[END_ACTIVITIES])
    num_start_activities_dev = len(conf_results[START_ACTIVITIES])
    num_end_activities_dev = len(conf_results[END_ACTIVITIES])
    footprints = conf_results[FOOTPRINTS_KEY]

    if dfg:
        sum_dfg = float(sum(x for x in dfg.values()))
        sum_dev = float(sum(dfg[x] for x in footprints))

        fitness = ((1.0 - sum_dev / sum_dfg) * (num_sequence_log + num_parallel_log) + (
                    num_start_activities_log + num_end_activities_log - num_start_activities_dev - num_end_activities_dev)) / (
                           num_sequence_log + num_parallel_log + num_start_activities_log + num_end_activities_log)
    else:
        # return fitness 1.0 if DFG is empty
        fitness = 1.0

    if fit_traces is not None:
        return {"perc_fit_traces": fit_traces, "log_fitness": fitness}

    return fitness


def fp_precision(fp_log, fp_model, parameters=None):
    """
    Calculates the footprints based precision provided the two footprints
    of the log and the model.

    Parameters
    --------------
    fp_log
        Footprints of the log
    fp_model
        Footprints of the model
    parameters
        Parameters of the algorithm

    Returns
    -------------
    precision
        Precision value (between 0 and 1)
    """
    if parameters is None:
        parameters = {}

    fp_log = flatten_fp(fp_log)
    fp_model = flatten_fp(fp_model)

    log_configurations = fp_log[Outputs.SEQUENCE.value].union(fp_log[Outputs.PARALLEL.value])
    model_configurations = fp_model[Outputs.SEQUENCE.value].union(fp_model[Outputs.PARALLEL.value])

    if model_configurations:
        return float(len(log_configurations.intersection(model_configurations))) / float(len(model_configurations))

    # return precision 1.0 if model configurations are empty
    return 1.0


def flatten_fp(fp: List[Dict[str, Any]]) -> Dict[str, Any]:
    """
    Flattens the trace-based footprints to the footprints of the overall log

    Parameters
    ---------------
    fp
        Trace-based footprints

    Returns
    --------------
    log_fp
        Overall log footprints
    """
    if isinstance(fp, list):
        res = {DFG: Counter(), SEQUENCE: set(), PARALLEL: set(), START_ACTIVITIES: set(), END_ACTIVITIES: set()}
        for el in fp:
            for x, y in el[DFG].items():
                res[DFG][x] += y
            res[SEQUENCE] = res[SEQUENCE].union(el[SEQUENCE])
            res[PARALLEL] = res[PARALLEL].union(el[PARALLEL])
            res[START_ACTIVITIES] = res[START_ACTIVITIES].union(el[START_ACTIVITIES])
            res[END_ACTIVITIES] = res[END_ACTIVITIES].union(el[END_ACTIVITIES])
        return res
    return fp


def flatten_conf(conf: List[Dict[str, Any]]) -> Dict[str, Any]:
    """
    Flattens the trace-based conformance checking results (obtained using footprints) to the conformance checking
    results on the overall log

    Parameters
    ----------------
    conf
        Trace-based conformance checking results

    Returns
    ----------------
    log_conf
        Overall log conformance checking results
    """
    if isinstance(conf, list):
        res = {FOOTPRINTS_KEY: set(), START_ACTIVITIES: set(), END_ACTIVITIES: set()}
        for el in conf:
            res[FOOTPRINTS_KEY] = res[FOOTPRINTS_KEY].union(el[FOOTPRINTS_KEY])
            res[START_ACTIVITIES] = res[START_ACTIVITIES].union(el[START_ACTIVITIES])
            res[END_ACTIVITIES] = res[END_ACTIVITIES].union(el[END_ACTIVITIES])
        return res
    return conf



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import pandas as pd
from typing import Optional, Dict, Any, Union
from pm4py.objects.ocel.obj import OCEL


def apply(ocel: OCEL, parameters: Optional[Dict[Any, Any]] = None) -> str:
    """
    Provides a string containing the required process mining domain knowledge for object-centric process mining structures
    (in order for the LLM to produce meaningful queries).

    Parameters
    ---------------
    ocel
        OCEL (2.0) object
    parameters
        Optional parameters of the method

    Returns
    --------------
    pm_knowledge
        String containing the required process mining knowledge
    """
    if parameters is None:
        parameters = {}

    descr = """
If you need to compute the duration of a lifecycle of an object, compute the difference between the timestamp of the last and the first event of the lifecycle.
If you need to compute the variant for an object, aggregate the names of the activities.
    """

    return descr



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import pandas as pd
from pm4py.util import exec_utils, constants, xes_constants, pandas_utils
from enum import Enum
from typing import Optional, Dict, Any


class Parameters(Enum):
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY
    LEFT_SUFFIX = "left_suffix"
    RIGHT_SUFFIX = "right_suffix"


def apply(dataframe1: pd.DataFrame, dataframe2: pd.DataFrame, case_relations: pd.DataFrame,
          parameters: Optional[Dict[Any, Any]] = None) -> pd.DataFrame:
    """
    Merges the two dataframes (dataframe1 and dataframe2), inserting the events of the second
    dataframe inside the cases of the first dataframe.
    This is done using a background knowledge provided in the case_relations dataframe, where the cases of the two dataframes
    are put in relations.
    E.g., if in dataframe1 and dataframe2 there are two case ID columns (case:concept:name),
    they are put in relations by case_relations having two columns case:concept:name_LEFT and case:concept:name_RIGHT

    Parameters
    -----------------
    dataframe1
        Reference dataframe (in which the events of the other dataframe are inserted)
    dataframe2
        Second dataframe (to insert in the cases of the first)
    case_relations
        Case relations dataframe
    parameters
        Parameters of the algorithm, including:
        - Parameters.TIMESTAMP_KEY => the timestamp key to use (default: time:timestamp)
        - Parameters.CASE_ID_KEY => the case identifier to use (default: case:concept:name)
        - Parameters.LEFT_SUFFIX => the suffix to the case identifier column of the first dataframe
        - Parameters.RIGHT_SUFFIX => the suffix to the case identifier column of the second dataframe

    Returns
    ----------------
    merged_dataframe
        Merged dataframe
    """
    if parameters is None:
        parameters = {}

    timestamp_key = exec_utils.get_param_value(Parameters.TIMESTAMP_KEY, parameters,
                                               xes_constants.DEFAULT_TIMESTAMP_KEY)
    left_suffix = exec_utils.get_param_value(Parameters.LEFT_SUFFIX, parameters, "_LEFT")
    right_suffix = exec_utils.get_param_value(Parameters.RIGHT_SUFFIX, parameters, "_RIGHT")
    case_id_key = exec_utils.get_param_value(Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME)

    dataframe2 = dataframe2.merge(case_relations, left_on=case_id_key, right_on=case_id_key + right_suffix)
    dataframe2[case_id_key] = dataframe2[case_id_key + left_suffix]

    dataframe1 = pandas_utils.concat([dataframe1, dataframe2])
    dataframe1 = dataframe1.sort_values([case_id_key, timestamp_key])

    return dataframe1



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import logging
from pm4py.streaming.algo.interface import StreamingAlgorithm


class DeclareStreamingConformance(StreamingAlgorithm):
    """
    Streaming Conformance Checking Algorithm for DECLARE models.
    Attempts to implement state-based checks for all Declare constraint types.
    When a violation occurs, prints out which constraints are violated.

    Implementation of:
    Maggi, Fabrizio Maria, et al. "Monitoring business constraints with linear temporal logic: An approach based on colored automata." Business Process Management: 9th International Conference, BPM 2011, Clermont-Ferrand, France, August 30-September 2, 2011. Proceedings 9. Springer Berlin Heidelberg, 2011.
    """

    def __init__(self, declare_model, parameters=None):
        super().__init__(parameters=parameters)
        self.declare_model = declare_model
        self._cases = {}
        self._total_events = 0
        self._total_deviations = 0
        self._deviations_per_time = []
        # Parse and build automata for all constraints
        self._constraints = self._parse_declare_model(declare_model)

    def _parse_declare_model(self, model):
        constraints = {}
        for template_type, activities_dict in model.items():
            constraints[template_type] = {}
            for activities, _params in activities_dict.items():
                if isinstance(activities, str):
                    activities = (activities,)
                constraints[template_type][activities] = self._create_automaton_for_constraint(template_type,
                                                                                               activities)
        return constraints

    # Automaton construction methods start here

    def _dummy_automaton(self):
        return {
            "initial_state": {"name": "init"},
            "states": {
                "init": {"on_event": lambda a, e, s: ("init", False)}
            }
        }

    def _existence_automaton(self, A):
        def on_event_init(a, e, s):
            if e.get("concept:name") == A:
                return ("seen", False)
            return ("init", False)

        def on_event_seen(a, e, s):
            return ("seen", False)

        return {
            "initial_state": {"name": "init"},
            "states": {
                "init": {"on_event": on_event_init},
                "seen": {"on_event": on_event_seen}
            }
        }

    def _absence_automaton(self, A):
        def on_event_init(a, e, s):
            if e.get("concept:name") == A:
                return ("violated", True)
            return ("init", False)

        return {
            "initial_state": {"name": "init"},
            "states": {
                "init": {"on_event": on_event_init},
                "violated": {"on_event": lambda a, e, s: ("violated", False)}
            }
        }

    def _exactly_one_automaton(self, A):
        def on_event_init(a, e, s):
            if e.get("concept:name") == A:
                return ("seen_once", False)
            return ("init", False)

        def on_event_seen_once(a, e, s):
            if e.get("concept:name") == A:
                return ("violated", True)
            return ("seen_once", False)

        return {
            "initial_state": {"name": "init"},
            "states": {
                "init": {"on_event": on_event_init},
                "seen_once": {"on_event": on_event_seen_once},
                "violated": {"on_event": lambda a, e, s: ("violated", False)}
            }
        }

    def _init_automaton(self, A):
        def on_event_init(a, e, s):
            if s.get("received_first", False) is False:
                s["received_first"] = True
                if e.get("concept:name") != A:
                    return ("violated", True)
                else:
                    return ("ok", False)
            else:
                return ("ok", False)

        return {
            "initial_state": {"name": "init", "received_first": False},
            "states": {
                "init": {"on_event": on_event_init},
                "ok": {"on_event": lambda a, e, s: ("ok", False)},
                "violated": {"on_event": lambda a, e, s: ("violated", False)}
            }
        }

    def _responded_existence_automaton(self, A, B):
        def on_event_init(a, e, s):
            act = e.get("concept:name")
            if act == A:
                s["A_seen"] = True
            if act == B:
                s["B_seen"] = True
            # No immediate violation, since we can't confirm end of trace
            return ("init", False)

        return {
            "initial_state": {"name": "init", "A_seen": False, "B_seen": False},
            "states": {"init": {"on_event": on_event_init}}
        }

    def _coexistence_automaton(self, A, B):
        def on_event_init(a, e, s):
            act = e.get("concept:name")
            if act == A:
                s["A_seen"] = True
            if act == B:
                s["B_seen"] = True
            return ("init", False)

        return {
            "initial_state": {"name": "init", "A_seen": False, "B_seen": False},
            "states": {"init": {"on_event": on_event_init}}
        }

    def _response_automaton(self, A, B):
        def on_event_init(a, e, s):
            act = e.get("concept:name")
            pA = s.get("pending_As", 0)
            if act == A:
                pA += 1
            if act == B and pA > 0:
                pA -= 1
            s["pending_As"] = pA
            return ("init", False)

        return {
            "initial_state": {"name": "init", "pending_As": 0},
            "states": {"init": {"on_event": on_event_init}}
        }

    def _precedence_automaton(self, A, B):
        def on_event_init(a, e, s):
            act = e.get("concept:name")
            if act == B and not s.get("A_occurred", False):
                return ("violated", True)
            if act == A:
                s["A_occurred"] = True
            return ("init", False)

        return {
            "initial_state": {"name": "init", "A_occurred": False},
            "states": {
                "init": {"on_event": on_event_init},
                "violated": {"on_event": lambda a, e, s: ("violated", False)}
            }
        }

    def _succession_automaton(self, A, B):
        def on_event_init(a, e, s):
            act = e.get("concept:name")
            pA = s.get("pending_As", 0)
            if act == B and not s.get("A_seen", False):
                return ("violated", True)
            if act == A:
                s["A_seen"] = True
                pA += 1
            elif act == B and pA > 0:
                pA -= 1
            s["pending_As"] = pA
            return ("init", False)

        return {
            "initial_state": {"name": "init", "A_seen": False, "pending_As": 0},
            "states": {
                "init": {"on_event": on_event_init},
                "violated": {"on_event": lambda a, e, s: ("violated", False)}
            }
        }

    def _altresponse_automaton(self, A, B):
        def on_event_init(a, e, s):
            act = e.get("concept:name")
            if act == A:
                if s.get("waiting_for_B", False):
                    return ("violated", True)
                s["waiting_for_B"] = True
            if act == B and s.get("waiting_for_B", False):
                s["waiting_for_B"] = False
            return ("init", False)

        return {
            "initial_state": {"name": "init", "waiting_for_B": False},
            "states": {
                "init": {"on_event": on_event_init},
                "violated": {"on_event": lambda a, e, s: ("violated", False)}
            }
        }

    def _altprecedence_automaton(self, A, B):
        def on_event_init(a, e, s):
            act = e.get("concept:name")
            if act == B:
                if s.get("waiting_for_A", True):
                    return ("violated", True)
                s["waiting_for_A"] = True
            if act == A:
                s["waiting_for_A"] = False
            return ("init", False)

        return {
            "initial_state": {"name": "init", "waiting_for_A": True},
            "states": {
                "init": {"on_event": on_event_init},
                "violated": {"on_event": lambda a, e, s: ("violated", False)}
            }
        }

    def _altsuccession_automaton(self, A, B):
        def on_event_init(a, e, s):
            act = e.get("concept:name")
            wB = s.get("waiting_for_B", False)
            wA = s.get("waiting_for_A", True)
            if act == A:
                if wB:
                    return ("violated", True)
                wB = True
                wA = False
            elif act == B:
                if wA:
                    return ("violated", True)
                wA = True
                wB = False
            s["waiting_for_A"] = wA
            s["waiting_for_B"] = wB
            return ("init", False)

        return {
            "initial_state": {"name": "init", "waiting_for_B": False, "waiting_for_A": True},
            "states": {
                "init": {"on_event": on_event_init},
                "violated": {"on_event": lambda a, e, s: ("violated", False)}
            }
        }

    def _chainresponse_automaton(self, A, B):
        def on_event_init(a, e, s):
            act = e.get("concept:name")
            expB = s.get("expecting_B", False)
            if expB:
                if act != B:
                    return ("violated", True)
                s["expecting_B"] = False
            if act == A:
                s["expecting_B"] = True
            return ("init", False)

        return {
            "initial_state": {"name": "init", "expecting_B": False},
            "states": {
                "init": {"on_event": on_event_init},
                "violated": {"on_event": lambda a, e, s: ("violated", False)}
            }
        }

    def _chainprecedence_automaton(self, A, B):
        def on_event_init(a, e, s):
            act = e.get("concept:name")
            last = s.get("last", None)
            if act == B and last != A:
                return ("violated", True)
            s["last"] = act
            return ("init", False)

        return {
            "initial_state": {"name": "init", "last": None},
            "states": {
                "init": {"on_event": on_event_init},
                "violated": {"on_event": lambda a, e, s: ("violated", False)}
            }
        }

    def _chainsuccession_automaton(self, A, B):
        def on_event_init(a, e, s):
            act = e.get("concept:name")
            expB = s.get("expecting_B", False)
            last = s.get("last", None)
            if expB:
                if act != B:
                    return ("violated", True)
                expB = False
            if act == B and last != A:
                return ("violated", True)
            if act == A:
                expB = True
            s["expecting_B"] = expB
            s["last"] = act
            return ("init", False)

        return {
            "initial_state": {"name": "init", "expecting_B": False, "last": None},
            "states": {
                "init": {"on_event": on_event_init},
                "violated": {"on_event": lambda a, e, s: ("violated", False)}
            }
        }

    def _noncoexistence_automaton(self, A, B):
        def on_event_init(a, e, s):
            act = e.get("concept:name")
            if act == A:
                s["A_seen"] = True
                if s.get("B_seen", False):
                    return ("violated", True)
            if act == B:
                s["B_seen"] = True
                if s.get("A_seen", False):
                    return ("violated", True)
            return ("init", False)

        return {
            "initial_state": {"name": "init", "A_seen": False, "B_seen": False},
            "states": {
                "init": {"on_event": on_event_init},
                "violated": {"on_event": lambda a, e, s: ("violated", False)}
            }
        }

    def _nonsuccession_automaton(self, A, B):
        def on_event_init(a, e, s):
            act = e.get("concept:name")
            if act == A:
                s["A_occurred"] = True
            if act == B and s.get("A_occurred", False):
                return ("violated", True)
            return ("init", False)

        return {
            "initial_state": {"name": "init", "A_occurred": False},
            "states": {
                "init": {"on_event": on_event_init},
                "violated": {"on_event": lambda a, e, s: ("violated", False)}
            }
        }

    def _nonchainsuccession_automaton(self, A, B):
        def on_event_init(a, e, s):
            act = e.get("concept:name")
            lA = s.get("last_was_A", False)
            if lA and act == B:
                return ("violated", True)
            lA = (act == A)
            s["last_was_A"] = lA
            return ("init", False)

        return {
            "initial_state": {"name": "init", "last_was_A": False},
            "states": {
                "init": {"on_event": on_event_init},
                "violated": {"on_event": lambda a, e, s: ("violated", False)}
            }
        }

    def _create_automaton_for_constraint(self, template_type, activities):
        if template_type == "existence":
            return self._existence_automaton(activities[0])
        elif template_type == "absence":
            return self._absence_automaton(activities[0])
        elif template_type == "exactly_one":
            return self._exactly_one_automaton(activities[0])
        elif template_type == "init":
            return self._init_automaton(activities[0])
        elif template_type == "responded_existence":
            return self._responded_existence_automaton(activities[0], activities[1])
        elif template_type == "coexistence":
            return self._coexistence_automaton(activities[0], activities[1])
        elif template_type == "response":
            return self._response_automaton(activities[0], activities[1])
        elif template_type == "precedence":
            return self._precedence_automaton(activities[0], activities[1])
        elif template_type == "succession":
            return self._succession_automaton(activities[0], activities[1])
        elif template_type == "altresponse":
            return self._altresponse_automaton(activities[0], activities[1])
        elif template_type == "altprecedence":
            return self._altprecedence_automaton(activities[0], activities[1])
        elif template_type == "altsuccession":
            return self._altsuccession_automaton(activities[0], activities[1])
        elif template_type == "chainresponse":
            return self._chainresponse_automaton(activities[0], activities[1])
        elif template_type == "chainprecedence":
            return self._chainprecedence_automaton(activities[0], activities[1])
        elif template_type == "chainsuccession":
            return self._chainsuccession_automaton(activities[0], activities[1])
        elif template_type == "noncoexistence":
            return self._noncoexistence_automaton(activities[0], activities[1])
        elif template_type == "nonsuccession":
            return self._nonsuccession_automaton(activities[0], activities[1])
        elif template_type == "nonchainsuccession":
            return self._nonchainsuccession_automaton(activities[0], activities[1])
        else:
            return self._dummy_automaton()

    def _process(self, event):
        case_id = event.get("case:concept:name", "undefined_case")
        self._total_events += 1

        if case_id not in self._cases:
            case_data = {
                "constraints_state": {},
                "deviations": 0,
                "events": 0
            }
            for template_type, constraints_dict in self._constraints.items():
                for activities, automaton in constraints_dict.items():
                    state_data = dict(automaton["initial_state"])
                    case_data["constraints_state"][(template_type, activities)] = (state_data["name"], state_data)
            self._cases[case_id] = case_data

        self._cases[case_id]["events"] += 1
        current_case_data = self._cases[case_id]
        deviations_in_this_event = 0
        violated_constraints = []

        for (template_type, activities), (state_name, state_data) in current_case_data["constraints_state"].items():
            automaton = self._constraints[template_type][activities]
            current_state = automaton["states"][state_name]
            on_event = current_state["on_event"]
            new_state_name, violated = on_event(automaton, event, state_data)
            current_case_data["constraints_state"][(template_type, activities)] = (new_state_name, state_data)
            if violated:
                deviations_in_this_event += 1
                violated_constraints.append((template_type, activities))

        if deviations_in_this_event > 0:
            current_case_data["deviations"] += deviations_in_this_event
            self._total_deviations += deviations_in_this_event
            # Print types of violated constraints
            violated_types = [vt for vt, _acts in violated_constraints]
            logging.error(
                f"Case {case_id} - Deviations detected: {deviations_in_this_event}. Violated constraint types: {violated_types}")

        timestamp = event.get("time:timestamp", self._total_events)
        self._deviations_per_time.append((timestamp, deviations_in_this_event))

    def _current_result(self):
        result = {
            "total_events_processed": self._total_events,
            "total_deviations": self._total_deviations,
            "deviations_per_time": self._deviations_per_time,
            "cases": {}
        }
        for c_id, c_data in self._cases.items():
            constraints_state = {}
            for k, (st_name, st_data) in c_data["constraints_state"].items():
                constraints_state[str(k)] = st_name
            result["cases"][c_id] = {
                "events": c_data["events"],
                "deviations": c_data["deviations"],
                "constraints_state": constraints_state
            }
        return result


def apply(declare_model, parameters=None):
    return DeclareStreamingConformance(declare_model, parameters)



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from typing import Union, Optional, Dict, Any
import pandas as pd
from sqlite3 import Connection as SQ3_Connection
from pm4py.objects.ocel.obj import OCEL
from pm4py.util import pandas_utils, exec_utils
from pm4py.algo.querying.llm.injection.db_knowledge.variants import pandas_duckdb, sqlite3_traditional


def apply(db: Union[pd.DataFrame, SQ3_Connection, OCEL], variant=None, parameters: Optional[Dict[Any, Any]] = None) -> str:
    """
    Provides a string containing the required database knowledge for database querying
    (in order for the LLM to produce meaningful queries).

    Parameters
    ---------------
    db
        Database
    variant
        Variant of the method to be used (pandas_duckdb, sqlite3_traditional)
    parameters
        Variant-specific parameters

    Returns
    ---------------
    db_knowledge
        String containing the required database knowledge.
    """
    if parameters is None:
        parameters = {}

    if variant is None:
        if pandas_utils.check_is_pandas_dataframe(db):
            variant = pandas_duckdb
        elif isinstance(db, SQ3_Connection):
            variant = sqlite3_traditional

    if variant is None:
        return "\n\n"

    return exec_utils.get_variant(variant).apply(db, parameters)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or 
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from typing import List, Optional, Dict, Any, Tuple

from pm4py.algo.discovery.powl.inductive.cuts.factory import T, CutFactory
from pm4py.algo.discovery.powl.inductive.cuts.loop import POWLLoopCutUVCL
from pm4py.algo.discovery.powl.inductive.cuts.xor import POWLExclusiveChoiceCutUVCL
from pm4py.algo.discovery.powl.inductive.variants.dynamic_clustering.dynamic_clustering_partial_order_cut import \
    DynamicClusteringPartialOrderCutUVCL
from pm4py.algo.discovery.inductive.dtypes.im_ds import IMDataStructure
from pm4py.objects.powl.obj import POWL
from pm4py.objects.dfg import util as dfu


class CutFactoryPOWLDynamicClustering(CutFactory):

    @classmethod
    def get_cuts(cls, obj, parameters=None):
        return [POWLExclusiveChoiceCutUVCL, POWLLoopCutUVCL, DynamicClusteringPartialOrderCutUVCL]

    @classmethod
    def find_cut(cls, obj: IMDataStructure, parameters: Optional[Dict[str, Any]] = None) -> Optional[
            Tuple[POWL, List[T]]]:
        alphabet = sorted(dfu.get_vertices(obj.dfg), key=lambda g: g.__str__())
        if len(alphabet) < 2:
            return None
        for c in CutFactoryPOWLDynamicClustering.get_cuts(obj):
            r = c.apply(obj, parameters)
            if r is not None:
                return r
        return None



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from typing import Optional, Dict, Any, Union, List

import numpy as np
import pandas as pd

from pm4py.util import exec_utils, constants, xes_constants


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY
    SUBCASE_CONCAT_STR = "subcase_concat_str"
    CUT_MODE = "cut_mode"


def apply(df: pd.DataFrame, activity: Union[str, List[str]],
          parameters: Optional[Dict[Union[str, Parameters], Any]] = None) -> pd.DataFrame:
    """
    Splits the cases of a log (Pandas dataframe) into subcases based on the provision of an activity.
    There are as many subcases as many occurrences of a given activity occur.

    Example:
    Original log:

    [['register request', 'examine casually', 'check ticket', 'decide', 'reinitiate request', 'examine thoroughly',
    'check ticket', 'decide', 'pay compensation'],
    ['register request', 'examine casually', 'check ticket', 'decide', 'reinitiate request', 'check ticket',
    'examine casually', 'decide', 'reinitiate request', 'examine casually', 'check ticket', 'decide', 'reject request']]


    Log filtered using the activity split filter on 'reinitiate request' with cut_mode='this':

    [['register request', 'examine casually', 'check ticket', 'decide'],
    ['reinitiate request', 'examine thoroughly', 'check ticket', 'decide', 'pay compensation'],
    ['register request', 'examine casually', 'check ticket', 'decide'],
    ['reinitiate request', 'check ticket', 'examine casually', 'decide'],
    ['reinitiate request', 'examine casually', 'check ticket', 'decide', 'reject request']]


    Log filtered using the activity split filter on 'reinitiate request' with cut_mode='next':

    [['register request', 'examine casually', 'check ticket', 'decide', 'reinitiate request'],
    ['examine thoroughly', 'check ticket', 'decide', 'pay compensation'],
    ['register request', 'examine casually', 'check ticket', 'decide', 'reinitiate request'],
    ['check ticket', 'examine casually', 'decide', 'reinitiate request'],
    ['examine casually', 'check ticket', 'decide', 'reject request']]


    Parameters
    ----------------
    df
        Dataframe
    activity
        Activity (or collection of activities)
    parameters
        Parameters of the algorithm, including:
        - Parameters.ACTIVITY_KEY => activity key
        - Parameters.CASE_ID_KEY => case id
        - Parameters.SUBCASE_CONCAT_STR => concatenator between the case id and the subtrace index in the filtered df
        - Parameters.CUT_MODE => mode of cut:
            - "this" means that an event with the specified activity goes to the next subcase
            - "next" means that the following event (to the given activity) goes to the next subcase.

    Returns
    ----------------
    filtered_df
        Dataframe in which the cases are split into subcases

    """
    if parameters is None:
        parameters = {}

    activity_key = exec_utils.get_param_value(Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY)
    case_id_key = exec_utils.get_param_value(Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME)
    subcase_concat_str = exec_utils.get_param_value(Parameters.SUBCASE_CONCAT_STR, parameters, "##@@")
    cut_mode = exec_utils.get_param_value(Parameters.CUT_MODE, parameters, "this")

    act_equal = lambda x: x == activity if type(activity) is str else x in activity

    df = df.copy()
    cases = df[case_id_key].to_numpy()
    activities = df[activity_key].to_numpy()
    c_unq, c_ind, c_counts = np.unique(cases, return_index=True, return_counts=True)
    res = []

    i = 0
    while i < len(c_unq):
        rel_count = 0
        this_case = str(c_unq[i]) + subcase_concat_str + str(rel_count)

        j = 0
        while j < c_counts[i]:
            if act_equal(activities[c_ind[i] + j]):
                rel_count += 1
                next_case = str(c_unq[i]) + subcase_concat_str + str(rel_count)
                if cut_mode == "this":
                    res.append(next_case)
                else:
                    res.append(this_case)
                this_case = next_case
            else:
                res.append(this_case)
            j = j + 1
        i = i + 1

    df[case_id_key] = res

    return df



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.objects.powl.obj import POWL
from typing import Optional, Dict, Any
from copy import deepcopy


def apply(powl: POWL, string_dictio: Dict[str, str], rec_depth=0, parameters: Optional[Dict[Any, Any]] = None) -> POWL:
    """
    Replaces the labels in the given POWL object using the provided dictionary.

    Parameters
    ---------------
    powl
        POWL
    string_dictio
        Correspondence dictionary (old labels -> new labels)

    Returns
    ----------------
    revised_powl
        Revised POWL
    """
    if parameters is None:
        parameters = {}

    if rec_depth == 0:
        powl = deepcopy(powl)

    if powl.label is not None and powl.label in string_dictio:
        powl.label = string_dictio[powl.label]

    for child in powl.children:
        apply(child, string_dictio, rec_depth=rec_depth+1, parameters=parameters)

    return powl



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import time
from enum import Enum
from typing import Optional, Dict, Any

import polars as pl

from pm4py.objects.dfg.obj import DFG
from pm4py.util import constants, exec_utils
from pm4py.util import xes_constants as xes_util


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY


CONST_AUX_ACT = 'aux_act_'
CONST_AUX_CASE = 'aux_case_'
CONST_COUNT = 'count_'


def apply(log: pl.DataFrame, parameters: Optional[Dict[str, Any]] = None) -> DFG:
    parameters = {} if parameters is None else parameters
    act_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_util.DEFAULT_NAME_KEY)
    cid_key = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, constants.CASE_ATTRIBUTE_GLUE)
    time_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY, parameters, xes_util.DEFAULT_TIMESTAMP_KEY)
    aux_act = CONST_AUX_ACT + str(time.time())
    aux_case = CONST_AUX_CASE + str(time.time())
    df = log[[cid_key, act_key, time_key]].clone()
    df = df.sort([cid_key, time_key])
    df = df[[cid_key, act_key]]
    df = df.with_column(df[act_key].shift(-1).alias(aux_act))
    df = df.with_column(df[cid_key].shift(-1).alias(aux_case))
    dfg = DFG()

    excl_starter = df[0, act_key]
    borders = df.filter(df[cid_key] != df[aux_case])

    for d in filter(lambda d: d[aux_act] is not None, borders.groupby([aux_act]).count().to_dicts()):
        v = d['count'] + 1 if d[aux_act] == excl_starter else d['count']
        dfg.start_activities[d[aux_act]] = v

    for d in filter(lambda d: d[act_key] is not None, borders.groupby([act_key]).count().to_dicts()):
        dfg.end_activities[d[act_key]] = d['count']

    for d in df.filter((df[cid_key] == df[aux_case])).groupby([act_key, aux_act]).count().to_dicts():
        dfg.graph[(d[act_key], d[aux_act])] = d['count']

    return dfg



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.algo.simulation.playout.process_tree.variants import basic_playout
from pm4py.algo.simulation.playout.process_tree.variants import extensive, topbottom
from enum import Enum
from pm4py.util import exec_utils
from pm4py.objects.process_tree.obj import ProcessTree
from typing import Optional, Dict, Any
from pm4py.objects.log.obj import EventLog


class Variants(Enum):
    BASIC_PLAYOUT = basic_playout
    EXTENSIVE = extensive
    TOPBOTTOM = topbottom


DEFAULT_VARIANT = Variants.TOPBOTTOM


def apply(tree: ProcessTree, variant=DEFAULT_VARIANT, parameters: Optional[Dict[Any, Any]] = None) -> EventLog:
    """
    Performs a playout of a process tree

    Parameters
    ---------------
    tree
        Process tree
    variant
        Variant of the algorithm:
        - Variants.BASIC_PLAYOUT: basic playout
        - Variants.EXTENSIVE: extensive playout (all the possible traces)
    parameters
        Parameters of the algorithm
    """
    if parameters is None:
        parameters = {}

    return exec_utils.get_variant(variant).apply(tree, parameters=parameters)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from pm4py.util import exec_utils
from pm4py.objects.ocel import constants as ocel_constants
from collections import Counter
from pm4py.objects.ocel.obj import OCEL
from typing import Optional, Dict, Any


class Parameters(Enum):
    EVENT_ID = ocel_constants.PARAM_EVENT_ID
    OBJECT_ID = ocel_constants.PARAM_OBJECT_ID
    OBJECT_TYPE = ocel_constants.PARAM_OBJECT_TYPE


def get_objects_ot_count(ocel: OCEL, parameters: Optional[Dict[Any, Any]] = None) -> Dict[str, Dict[str, int]]:
    """
    Counts for each event the number of related objects per type

    Parameters
    -------------------
    ocel
        Object-centric Event log
    parameters
        Parameters of the algorithm, including:
        - Parameters.EVENT_ID => the event identifier to be used
        - Parameters.OBJECT_ID => the object identifier to be used
        - Parameters.OBJECT_TYPE => the object type to be used

    Returns
    -------------------
    dict_ot
        Dictionary associating to each event identifier a dictionary with the number of related objects
    """
    if parameters is None:
        parameters = {}

    event_id = exec_utils.get_param_value(Parameters.EVENT_ID, parameters, ocel.event_id_column)
    object_id = exec_utils.get_param_value(Parameters.OBJECT_ID, parameters, ocel.object_id_column)
    object_type = exec_utils.get_param_value(Parameters.OBJECT_TYPE, parameters, ocel.object_type_column)

    ref0 = ocel.relations.groupby(event_id)[object_type].agg(list).to_dict()
    ref = {x: Counter(y) for x, y in ref0.items()}

    return ref



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import numpy as np
from pm4py.util.lp import solver


def get_c_matrix(PS_matrix, duration_matrix, activities, activities_counter):
    """
    Calculates the C-matrix out of the PS matrix and the duration matrix

    Parameters
    --------------
    PS_matrix
        PS matrix
    duration_matrix
        Duration matrix
    activities
        Ordered list of activities of the log
    activities_counter
        Counter of activities

    Returns
    --------------
    c_matrix
        C matrix
    """
    C_matrix = np.zeros((len(activities), len(activities)))
    for i in range(len(activities)):
        for j in range(len(activities)):
            val = duration_matrix[i, j] / PS_matrix[i, j] * 1 / (
                min(activities_counter[activities[i]], activities_counter[activities[j]])) if PS_matrix[
                                                                                                  i, j] > 0 else 0
            if val == 0:
                val = 100000000000
            C_matrix[i, j] = val
    return C_matrix


def resolve_LP(C_matrix, duration_matrix, activities, activities_counter):
    """
    Formulates and solve the LP problem

    Parameters
    --------------
    C_matrix
        C_matrix
    duration_matrix
        Duration matrix
    activities
        Ordered list of activities of the log
    activities_counter
        Counter of activities

    Returns
    -------------
    dfg
        Directly-Follows Graph
    performance_dfg
        Performance DFG (containing the estimated performance for the arcs)
    """
    edges = [(i, j) for i in range(len(activities)) for j in range(len(activities))]
    c = [C_matrix[i, j] for i in range(len(activities)) for j in range(len(activities))]
    edges_sources = {i: [z for z in range(len(edges)) if edges[z][0] == i] for i in range(len(activities))}
    edges_targets = {j: [z for z in range(len(edges)) if edges[z][1] == j] for j in range(len(activities))}
    activities_occurrences = {i: activities_counter[activities[i]] for i in range(len(activities))}
    Aeq = []
    beq = []
    for i in range(len(activities)):
        rec = [0] * len(edges)
        for e in edges_sources[i]:
            rec[e] = 1
        Aeq.append(rec)
        beq.append(activities_occurrences[i])
    for j in range(len(activities)):
        rec = [0] * len(edges)
        for e in edges_targets[j]:
            rec[e] = 1
        Aeq.append(rec)
        beq.append(activities_occurrences[j])
    Aeq = np.asmatrix(Aeq).astype(np.float64)
    beq = np.asmatrix(beq).transpose().astype(np.float64)
    Aub = []
    bub = []
    for i in range(len(activities)):
        for e in edges_sources[i]:
            rec = [0] * len(edges)
            rec[e] = 1
            Aub.append(rec)
            bub.append(activities_occurrences[i])
            rec = [-x for x in rec]
            Aub.append(rec)
            bub.append(0)
    for j in range(len(activities)):
        for e in edges_targets[j]:
            rec = [0] * len(edges)
            rec[e] = 1
            Aub.append(rec)
            bub.append(activities_occurrences[j])
            rec = [-x for x in rec]
            Aub.append(rec)
            bub.append(0)
    Aub = np.asmatrix(Aub).astype(np.float64)
    bub = np.asmatrix(bub).transpose().astype(np.float64)

    use_cvxopt = False
    if solver.DEFAULT_LP_SOLVER_VARIANT == solver.CVXOPT_SOLVER_CUSTOM_ALIGN or solver.DEFAULT_LP_SOLVER_VARIANT == solver.CVXOPT_SOLVER_CUSTOM_ALIGN_ILP:
        use_cvxopt = True

    if use_cvxopt:
        from cvxopt import matrix

        c = matrix(c)
        Aub = matrix(Aub)
        bub = matrix(bub)
        Aeq = matrix(Aeq)
        beq = matrix(beq)

    res = solver.apply(c, Aub, bub, Aeq, beq, variant=solver.DEFAULT_LP_SOLVER_VARIANT)
    points = solver.get_points_from_sol(res, variant=solver.DEFAULT_LP_SOLVER_VARIANT)
    points = [round(p) for p in points]

    dfg = {}
    performance_dfg = {}

    for idx, p in enumerate(points):
        if p > 0:
            dfg[(activities[edges[idx][0]], activities[edges[idx][1]])] = p
            performance_dfg[(activities[edges[idx][0]], activities[edges[idx][1]])] = duration_matrix[
                edges[idx][0], edges[idx][1]]
    return dfg, performance_dfg


def match_return_avg_time(ai, aj, exact=False):
    """
    Matches two list of times (exact or greedy)
    and returns the average.

    Parameters
    --------------
    ai
        First list
    aj
        Second list

    Returns
    ---------------
    times_mean
        Mean of times
    """
    from statistics import mean

    if exact:
        from pm4py.statistics.util import times_bipartite_matching
        matching = times_bipartite_matching.exact_match_minimum_average(ai, aj)
        ret_exact = mean([x[1] - x[0] for x in matching]) if matching else 0
        return ret_exact
    else:
        ret_greedy = greedy_match_return_avg_time(ai, aj)
        return ret_greedy


def greedy_match_return_avg_time(ai, aj):
    """
    Matches two list of times with a greedy method
    and returns the average.

    Parameters
    --------------
    ai
        First list
    aj
        Second list
    parameters
        Parameters of the algorithm

    Returns
    ---------------
    times_mean
        Mean of times
    """
    from statistics import mean

    tm0 = calculate_time_match_fifo(ai, aj)
    td0 = mean([x[1] - x[0] for x in tm0]) if tm0 else 0
    tm1 = calculate_time_match_rlifo(ai, aj)
    td1 = mean([x[1] - x[0] for x in tm1]) if tm1 else 0
    return min(td0, td1)


def calculate_time_match_fifo(ai, aj, times0=None):
    """
    Associate the times between
    two lists of timestamps using FIFO

    Parameters
    --------------
    ai
        First list of timestamps
    aj
        Second list of timestamps
    times0
        Correspondence between execution times

    Returns
    --------------
    times0
        Correspondence between execution times
    """
    if times0 is None:
        times0 = []
    k = 0
    z = 0
    while k < len(ai):
        while z < len(aj):
            if ai[k] < aj[z]:
                times0.append((ai[k], aj[z]))
                z = z + 1
                break
            z = z + 1
        k = k + 1
    return times0


def calculate_time_match_rlifo(ai, aj, times1=None):
    """
    Associate the times between
    two lists of timestamps using LIFO (start from end)

    Parameters
    --------------
    ai
        First list of timestamps
    aj
        Second list of timestamps
    times0
        Correspondence between execution times

    Returns
    --------------
    times0
        Correspondence between execution times
    """
    if times1 is None:
        times1 = []
    k = len(ai) - 1
    z = len(aj) - 1
    while z >= 0:
        while k >= 0:
            if ai[k] < aj[z]:
                times1.append((ai[k], aj[z]))
                k = k - 1
                break
            k = k - 1
        z = z - 1
    return times1



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

EXISTENCE = "existence"
EXACTLY_ONE = "exactly_one"
INIT = "init"
RESPONDED_EXISTENCE = "responded_existence"
RESPONSE = "response"
PRECEDENCE = "precedence"
SUCCESSION = "succession"
ALTRESPONSE = "altresponse"
ALTPRECEDENCE = "altprecedence"
ALTSUCCESSION = "altsuccession"
CHAINRESPONSE = "chainresponse"
CHAINPRECEDENCE = "chainprecedence"
CHAINSUCCESSION = "chainsuccession"
ABSENCE = "absence"
COEXISTENCE = "coexistence"
NONCOEXISTENCE = "noncoexistence"
NONSUCCESSION = "nonsuccession"
NONCHAINSUCCESSION = "nonchainsuccession"



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from typing import Optional, Dict, Any, Union

import pandas as pd

from pm4py.util import constants, xes_constants, exec_utils


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY


INT_CASE_ACT_SIZE = "@@int_case_act_size"


def apply(df: pd.DataFrame, parameters: Optional[Dict[Union[str, Parameters], Any]] = None) -> Dict[str, int]:
    """
    Associates to each activity (with at least one rework) the number of cases in the log for which
    the rework happened.

    Parameters
    ------------------
    df
        Dataframe
    parameters
        Parameters of the algorithm, including:
        - Parameters.ACTIVITY_KEY => the attribute to be used as activity
        - Parameters.CASE_ID_KEY => the attribute to be used as case ID

    Returns
    ------------------
    dict
        Dictionary associating to each activity the number of cases for which the rework happened
    """
    if parameters is None:
        parameters = {}

    activity_key = exec_utils.get_param_value(Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY)
    case_id_key = exec_utils.get_param_value(Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME)

    df = df.copy()
    df = df[list({activity_key, case_id_key})]
    df[INT_CASE_ACT_SIZE] = df.groupby([activity_key, case_id_key]).cumcount()
    df = df[df[INT_CASE_ACT_SIZE] > 0]
    df = df.groupby([activity_key, case_id_key]).last()
    ret = df.groupby(activity_key).size().to_dict()

    return ret



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or 
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.algo.evaluation.generalization.variants import token_based
from enum import Enum
from pm4py.util import exec_utils
from typing import Optional, Dict, Any, Union
from pm4py.objects.log.obj import EventLog, EventStream
from pm4py.objects.petri_net.obj import PetriNet, Marking
import pandas as pd


class Variants(Enum):
    GENERALIZATION_TOKEN = token_based


GENERALIZATION_TOKEN = Variants.GENERALIZATION_TOKEN
VERSIONS = {GENERALIZATION_TOKEN}


def apply(log: Union[EventLog, EventStream, pd.DataFrame], petri_net: PetriNet, initial_marking: Marking, final_marking: Marking, parameters: Optional[Dict[Any, Any]] = None, variant=GENERALIZATION_TOKEN) -> float:
    if parameters is None:
        parameters = {}

    return exec_utils.get_variant(variant).apply(log,
                                                 petri_net,
                                                 initial_marking, final_marking, parameters=parameters)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

from enum import Enum
from pm4py.util import exec_utils, xes_constants
import pandas as pd
from typing import Optional, Dict, Any, Union, Tuple


class Parameters(Enum):
    ACTIVITY_KEY = "activity_key"
    RETURN_MAPPING = "return_mapping"


def apply(dataframe: pd.DataFrame, parameters: Optional[Dict[Any, Any]] = None) -> Union[
    pd.DataFrame, Tuple[pd.DataFrame, Dict[str, str]]]:
    """
    Remap the activities in a dataframe using an augmented alphabet to minimize the size of the encoding

    Running example:

        import pm4py
        from pm4py.objects.log.util import activities_to_alphabet
        from pm4py.util import constants

        dataframe = pm4py.read_xes("tests/input_data/running-example.xes")
        renamed_dataframe = activities_to_alphabet.apply(dataframe, parameters={constants.PARAMETER_CONSTANT_ACTIVITY_KEY: "concept:name"})
        print(renamed_dataframe)

    Parameters
    --------------
    dataframe
        Pandas dataframe
    parameters
        Parameters of the method, including:
        - Parameters.ACTIVITY_KEY => attribute to be used as activity
        - Parameters.RETURN_MAPPING => (boolean) enables the returning the mapping dictionary (so the original activities can be re-constructed)

    Returns
    --------------
    ren_dataframe
        Pandas dataframe in which the activities have been remapped to the (augmented) alphabet
    inv_mapping
        (if required) Dictionary associating to every letter of the (augmented) alphabet the original activity
    """
    if parameters is None:
        parameters = {}

    activity_key = exec_utils.get_param_value(Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY)
    return_mapping = exec_utils.get_param_value(Parameters.RETURN_MAPPING, parameters, False)

    activities_count = list(dataframe[activity_key].value_counts().to_dict())
    remap_dict = {}
    for index, act in enumerate(activities_count):
        result = ''
        while index >= 0:
            result = chr((index % 26) + ord('A')) + result
            index = index // 26 - 1
        remap_dict[act] = result
    dataframe[activity_key] = dataframe[activity_key].map(remap_dict)
    if return_mapping:
        inverse_dct = {y: x for x, y in remap_dict.items()}
        return dataframe, inverse_dct
    return dataframe



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from typing import Optional, Dict, Any, List

import pandas as pd

from pm4py.objects.conversion.log import converter as log_converter
from pm4py.objects.log.obj import EventStream
from pm4py.util import constants
from pm4py.util import exec_utils
from pm4py.util import points_subset
from pm4py.util import xes_constants, pandas_utils
from pm4py.util.dt_parsing.variants import strpfromiso
import numpy as np
import random
import traceback


LEGACY_PARQUET_TP_REPLACER = "AAA"
LEGACY_PARQUET_CASECONCEPTNAME = "caseAAAconceptAAAname"


class Parameters(Enum):
    PARTITION_COLUMN = "partition_column"
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY
    CASE_PREFIX = constants.CASE_ATTRIBUTE_PREFIX
    CASE_ATTRIBUTES = "case_attributes"
    MANDATORY_ATTRIBUTES = "mandatory_attributes"
    MAX_NO_CASES = "max_no_cases"
    MIN_DIFFERENT_OCC_STR_ATTR = 5
    MAX_DIFFERENT_OCC_STR_ATTR = 50
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    PARAM_ARTIFICIAL_START_ACTIVITY = constants.PARAM_ARTIFICIAL_START_ACTIVITY
    PARAM_ARTIFICIAL_END_ACTIVITY = constants.PARAM_ARTIFICIAL_END_ACTIVITY
    INDEX_KEY = "index_key"
    CASE_INDEX_KEY = "case_index_key"
    USE_EXTREMES_TIMESTAMP = "use_extremes_timestamp"
    ADD_CASE_IDENTIFIER_COLUMN = "add_case_identifier_column"
    DETERMINISTIC = "deterministic"


def insert_partitioning(df, num_partitions, parameters=None):
    """
    Insert the partitioning in the specified dataframe

    Parameters
    -------------
    df
        Dataframe
    num_partitions
        Number of partitions
    parameters
        Parameters of the algorithm

    Returns
    -------------
    df
        Partitioned dataframe
    """
    if parameters is None:
        parameters = {}

    case_index_key = exec_utils.get_param_value(Parameters.CASE_INDEX_KEY, parameters, constants.DEFAULT_CASE_INDEX_KEY)
    partition_column = exec_utils.get_param_value(Parameters.PARTITION_COLUMN, parameters, "@@partitioning")

    if case_index_key not in df.columns:
        from pm4py.util import pandas_utils
        df = pandas_utils.insert_case_index(df, case_index_key)

    df[partition_column] = df[case_index_key] % num_partitions

    return df


def legacy_parquet_support(df, parameters=None):
    """
    For legacy support, Parquet files columns could not contain
    a ":" that has been arbitrarily replaced by a replacer string.
    This string substitutes the replacer to the :

    Parameters
    ---------------
    dataframe
        Dataframe
    parameters
        Parameters of the algorithm
    """
    if parameters is None:
        parameters = {}

    df.columns = [x.replace(LEGACY_PARQUET_TP_REPLACER, ":") for x in df.columns]

    return df


def table_to_stream(table, parameters=None):
    """
    Converts a Pyarrow table to an event stream

    Parameters
    ------------
    table
        Pyarrow table
    parameters
        Possible parameters of the algorithm
    """
    if parameters is None:
        parameters = {}

    dict0 = table.to_pydict()
    keys = list(dict0.keys())
    # for legacy format support
    if LEGACY_PARQUET_CASECONCEPTNAME in keys:
        for key in keys:
            dict0[key.replace(LEGACY_PARQUET_TP_REPLACER, ":")] = dict0.pop(key)

    stream = EventStream([dict(zip(dict0, i)) for i in zip(*dict0.values())])

    return stream


def table_to_log(table, parameters=None):
    """
    Converts a Pyarrow table to an event log

    Parameters
    ------------
    table
        Pyarrow table
    parameters
        Possible parameters of the algorithm
    """
    if parameters is None:
        parameters = {}

    stream = table_to_stream(table, parameters=parameters)

    return log_converter.apply(stream, parameters=parameters)


def convert_timestamp_columns_in_df(df, timest_format=None, timest_columns=None):
    """
    Convert all dataframe columns in a dataframe

    Parameters
    -----------
    df
        Dataframe
    timest_format
        (If provided) Format of the timestamp columns in the CSV file
    timest_columns
        Columns of the CSV that shall be converted into timestamp

    Returns
    ------------
    df
        Dataframe with timestamp columns converted

    """
    if timest_format is None:
        timest_format = constants.DEFAULT_TIMESTAMP_PARSE_FORMAT

    if timest_format is None:
        timest_format = "mixed"

    for col in df.columns:
        if timest_columns is None or col in timest_columns:
            if "obj" in str(df[col].dtype) or "str" in str(df[col].dtype):
                try:
                    df[col] = pandas_utils.dataframe_column_string_to_datetime(df[col], format=timest_format, utc=True)
                except:
                    try:
                        df[col] = pandas_utils.dataframe_column_string_to_datetime(df[col], format=timest_format, exact=False, utc=True)
                    except:
                        try:
                            df[col] = pandas_utils.dataframe_column_string_to_datetime(df[col], format=timest_format)
                        except:
                            pass

    for col in df.columns:
        if "date" in str(df[col].dtype) or "time" in str(df[col].dtype):
            df[col] = strpfromiso.fix_dataframe_column(df[col])

    return df


def sample_dataframe(df, parameters=None):
    """
    Sample a dataframe on a given number of cases

    Parameters
    --------------
    df
        Dataframe
    parameters
        Parameters of the algorithm, including:
        - Parameters.CASE_ID_KEY
        - Parameters.CASE_ID_TO_RETAIN

    Returns
    -------------
    sampled_df
        Sampled dataframe
    """
    if parameters is None:
        parameters = {}

    deterministic = exec_utils.get_param_value(Parameters.DETERMINISTIC, parameters, False)
    case_id_key = exec_utils.get_param_value(Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME)
    max_no_cases = exec_utils.get_param_value(Parameters.MAX_NO_CASES, parameters, 100)

    case_ids = pandas_utils.format_unique(df[case_id_key].unique())
    case_ids = list(case_ids)

    if not deterministic:
        random.shuffle(case_ids)

    case_id_to_retain = points_subset.pick_chosen_points_list(min(max_no_cases, len(case_ids)), case_ids)

    return df[df[case_id_key].isin(case_id_to_retain)]


def automatic_feature_selection_df(df, parameters=None):
    """
    Performs an automatic feature selection on dataframes,
    keeping the features useful for ML purposes

    Parameters
    ---------------
    df
        Dataframe
    parameters
        Parameters of the algorithm

    Returns
    ---------------
    featured_df
        Dataframe with only the features that have been selected
    """
    if parameters is None:
        parameters = {}

    case_id_key = exec_utils.get_param_value(Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME)
    timestamp_key = exec_utils.get_param_value(Parameters.TIMESTAMP_KEY, parameters, xes_constants.DEFAULT_TIMESTAMP_KEY)
    activity_key = exec_utils.get_param_value(Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY)

    mandatory_attributes = exec_utils.get_param_value(Parameters.MANDATORY_ATTRIBUTES, parameters,
                                                      set(df.columns).intersection(
                                                          {case_id_key, activity_key,
                                                           timestamp_key}))

    min_different_occ_str_attr = exec_utils.get_param_value(Parameters.MIN_DIFFERENT_OCC_STR_ATTR, parameters, 5)
    max_different_occ_str_attr = exec_utils.get_param_value(Parameters.MAX_DIFFERENT_OCC_STR_ATTR, parameters, 50)

    cols_dtypes = {x: str(df[x].dtype) for x in df.columns}
    other_attributes_to_retain = set()

    no_all_cases = df[case_id_key].nunique()
    for x, y in cols_dtypes.items():
        attr_df = df.dropna(subset=[x])
        this_cases = attr_df[case_id_key].nunique()

        # in any case, keep attributes that appears at least once per case
        if this_cases == no_all_cases:
            if "float" in y or "int" in y:
                # (as in the classic log version) retain always float/int attributes
                other_attributes_to_retain.add(x)
            elif "obj" in y or "str" in y:
                # (as in the classic log version) keep string attributes if they have enough variability, but not too much
                # (that would be hard to explain)
                unique_val_count = df[x].nunique()
                if min_different_occ_str_attr <= unique_val_count <= max_different_occ_str_attr:
                    other_attributes_to_retain.add(x)
            else:
                # not consider the attribute after this feature selection if it has other types (for example, date)
                pass

    attributes_to_retain = mandatory_attributes.union(other_attributes_to_retain)

    return df[list(attributes_to_retain)]


def select_number_column(df: pd.DataFrame, fea_df: pd.DataFrame, col: str,
                         case_id_key=constants.CASE_CONCEPT_NAME) -> pd.DataFrame:
    """
    Extract a column for the features dataframe for the given numeric attribute

    Parameters
    --------------
    df
        Dataframe
    fea_df
        Feature dataframe
    col
        Numeric column
    case_id_key
        Case ID key

    Returns
    --------------
    fea_df
        Feature dataframe (desidered output)
    """
    df = df.dropna(subset=[col]).groupby(case_id_key).last().reset_index()[[case_id_key, col]]
    fea_df = fea_df.merge(df, on=[case_id_key], how="left", suffixes=('', '_y'))
    fea_df[col] = fea_df[col].astype(np.float32)
    return fea_df


def select_string_column(df: pd.DataFrame, fea_df: pd.DataFrame, col: str,
                         case_id_key=constants.CASE_CONCEPT_NAME) -> pd.DataFrame:
    """
    Extract N columns (for N different attribute values; hotencoding) for the features dataframe for the given string attribute

    Parameters
    --------------
    df
        Dataframe
    fea_df
        Feature dataframe
    col
        String column
    case_id_key
        Case ID key

    Returns
    --------------
    fea_df
        Feature dataframe (desidered output)
    """
    vals = pandas_utils.format_unique(df[col].unique())
    for val in vals:
        if val is not None:
            filt_df_cases = pandas_utils.format_unique(df[df[col] == val][case_id_key].unique())
            new_col = col + "_" + val.encode('ascii', errors='ignore').decode('ascii').replace(" ", "")
            fea_df[new_col] = fea_df[case_id_key].isin(filt_df_cases)
            fea_df[new_col] = fea_df[new_col].astype(np.float32)
    return fea_df


def get_features_df(df: pd.DataFrame, list_columns: List[str],
                    parameters: Optional[Dict[Any, Any]] = None) -> pd.DataFrame:
    """
    Given a dataframe and a list of columns, performs an automatic feature extraction

    Parameters
    ---------------
    df
        Dataframe
    list_column
        List of column to consider in the feature extraction
    parameters
        Parameters of the algorithm, including:
        - Parameters.CASE_ID_KEY: the case ID

    Returns
    ---------------
    fea_df
        Feature dataframe (desidered output)
    """
    if parameters is None:
        parameters = {}

    case_id_key = exec_utils.get_param_value(Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME)
    add_case_identifier_column = exec_utils.get_param_value(Parameters.ADD_CASE_IDENTIFIER_COLUMN, parameters, False)

    fea_df = pandas_utils.instantiate_dataframe({case_id_key: sorted(pandas_utils.format_unique(df[case_id_key].unique()))})
    for col in list_columns:
        if "obj" in str(df[col].dtype) or "str" in str(df[col].dtype):
            fea_df = select_string_column(df, fea_df, col, case_id_key=case_id_key)
        elif "float" in str(df[col].dtype) or "int" in str(df[col].dtype):
            fea_df = select_number_column(df, fea_df, col, case_id_key=case_id_key)
    fea_df = fea_df.sort_values(case_id_key)
    if not add_case_identifier_column:
        del fea_df[case_id_key]

    return fea_df


def automatic_feature_extraction_df(df: pd.DataFrame, parameters: Optional[Dict[Any, Any]] = None) -> pd.DataFrame:
    """
    Performs an automatic feature extraction given a dataframe

    Parameters
    --------------
    df
        Dataframe
    parameters
        Parameters of the algorithm, including:
        - Parameters.CASE_ID_KEY: the case ID
        - Parameters.MIN_DIFFERENT_OCC_STR_ATTR
        - Parameters.MAX_DIFFERENT_OCC_STR_ATTR

    Returns
    --------------
    fea_df
        Dataframe with the features
    """
    if parameters is None:
        parameters = {}

    case_id_key = exec_utils.get_param_value(Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME)
    timestamp_key = exec_utils.get_param_value(Parameters.TIMESTAMP_KEY, parameters, xes_constants.DEFAULT_TIMESTAMP_KEY)

    fea_sel_df = automatic_feature_selection_df(df, parameters=parameters)
    columns = set(fea_sel_df.columns)

    if case_id_key in columns:
        columns.remove(case_id_key)

    if timestamp_key in columns:
        columns.remove(timestamp_key)

    return get_features_df(fea_sel_df, list(columns), parameters=parameters)


def insert_artificial_start_end(df0: pd.DataFrame, parameters: Optional[Dict[Any, Any]] = None) -> pd.DataFrame:
    """
    Inserts the artificial start/end activities in a Pandas dataframe

    Parameters
    ------------------
    df0
        Dataframe
    parameters
        Parameters of the algorithm, including:
        - Parameters.CASE_ID_KEY: the case identifier
        - Parameters.TIMESTAMP_KEY: the timestamp
        - Parameters.ACTIVITY_KEY: the activity

    Returns
    -----------------
    enriched_df
        Dataframe with artificial start/end activities
    """
    if parameters is None:
        parameters = {}

    case_id_key = exec_utils.get_param_value(Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME)
    timestamp_key = exec_utils.get_param_value(Parameters.TIMESTAMP_KEY, parameters, xes_constants.DEFAULT_TIMESTAMP_KEY)
    activity_key = exec_utils.get_param_value(Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY)
    use_extremes_timestamp = exec_utils.get_param_value(Parameters.USE_EXTREMES_TIMESTAMP, parameters, False)

    artificial_start_activity = exec_utils.get_param_value(Parameters.PARAM_ARTIFICIAL_START_ACTIVITY, parameters, constants.DEFAULT_ARTIFICIAL_START_ACTIVITY)
    artificial_end_activity = exec_utils.get_param_value(Parameters.PARAM_ARTIFICIAL_END_ACTIVITY, parameters, constants.DEFAULT_ARTIFICIAL_END_ACTIVITY)

    index_key = exec_utils.get_param_value(Parameters.INDEX_KEY, parameters, constants.DEFAULT_INDEX_KEY)

    df = df0.copy()
    df = pandas_utils.insert_index(df, index_key)
    df = df.sort_values([case_id_key, timestamp_key, index_key])

    start_df = df[[case_id_key, timestamp_key]].groupby(case_id_key).first().reset_index()
    end_df = df[[case_id_key, timestamp_key]].groupby(case_id_key).last().reset_index()
    # stability trick: remove 1ms from the artificial start activity timestamp, add 1ms to the artificial end activity timestamp
    if use_extremes_timestamp:
        start_df[timestamp_key] = pd.Timestamp.min
        end_df[timestamp_key] = pd.Timestamp.max
        start_df[timestamp_key] = start_df[timestamp_key].dt.tz_localize("utc")
        end_df[timestamp_key] = end_df[timestamp_key].dt.tz_localize("utc")
    else:
        start_df[timestamp_key] = start_df[timestamp_key] - pd.Timedelta("1 ms")
        end_df[timestamp_key] = end_df[timestamp_key] + pd.Timedelta("1 ms")

    start_df[activity_key] = artificial_start_activity
    end_df[activity_key] = artificial_end_activity

    df = pandas_utils.concat([start_df, df, end_df])
    df = pandas_utils.insert_index(df, index_key)
    df = df.sort_values([case_id_key, timestamp_key, index_key])

    df.attrs = df0.attrs

    return df


def dataframe_to_activity_case_table(df: pd.DataFrame, parameters: Optional[Dict[Any, Any]] = None):
    """
    Transforms a Pandas dataframe into:
    - an "activity" table, containing the events and their attributes
    - a "case" table, containing the cases and their attributes

    Parameters
    --------------
    df
        Dataframe
    parameters
        Parameters of the algorithm that should be used, including:
        - Parameters.CASE_ID_KEY => the column to be used as case ID (shall be included both in the activity table and the case table)
        - Parameters.CASE_PREFIX => if a list of attributes at the case level is not provided, then all the ones of the dataframe
                                    starting with one of these are considered.
        - Parameters.CASE_ATTRIBUTES => the attributes of the dataframe to be used as case columns

    Returns
    ---------------
    activity_table
        Activity table
    case_table
        Case table
    """
    if parameters is None:
        parameters = {}

    # make sure we start from a dataframe object
    df = log_converter.apply(df, variant=log_converter.Variants.TO_DATA_FRAME, parameters=parameters)

    case_id_key = exec_utils.get_param_value(Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME)
    case_id_prefix = exec_utils.get_param_value(Parameters.CASE_PREFIX, parameters, constants.CASE_ATTRIBUTE_PREFIX)

    case_attributes = exec_utils.get_param_value(Parameters.CASE_ATTRIBUTES, parameters, set([x for x in df.columns if x.startswith(case_id_prefix)]))
    event_attributes = set([x for x in df.columns if x not in case_attributes])

    activity_table = df[event_attributes.union({case_id_key})]
    case_table = df[case_attributes.union({case_id_key})].groupby(case_id_key).first().reset_index()

    return activity_table, case_table



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.objects.ocel import constants
from enum import Enum
from pm4py.objects.ocel.obj import OCEL
from typing import Optional, Dict, Any, Set
from pm4py.util import exec_utils, pandas_utils
import pandas as pd


class Parameters(Enum):
    EVENT_ID = constants.PARAM_EVENT_ID
    OBJECT_ID = constants.PARAM_OBJECT_ID
    OBJECT_TYPE = constants.PARAM_OBJECT_TYPE


def _get_unique_ids(df: pd.DataFrame, column: str) -> Set:
    """Helper function to get unique IDs from a DataFrame column."""
    if df.empty:
        return set()
    return set(pandas_utils.format_unique(df[column].unique()))


def propagate_event_filtering(
        ocel: OCEL, parameters: Optional[Dict[Any, Any]] = None
) -> OCEL:
    """
    Propagates the filtering at the event level to the remaining parts of the OCEL structure
    (objects, relations)

    Parameters
    ----------------
    ocel
        Object-centric event log
    parameters
        Parameters of the algorithm, including:
        - Parameters.EVENT_ID => the column to be used as case identifier
        - Parameters.OBJECT_ID => the column to be used as object identifier
        - Parameters.OBJECT_TYPE => the column to be used as object type

    Returns
    ----------------
    ocel
        Object-centric event log with propagated filter
    """
    if parameters is None:
        parameters = {}

    event_id = exec_utils.get_param_value(
        Parameters.EVENT_ID, parameters, ocel.event_id_column
    )
    object_id = exec_utils.get_param_value(
        Parameters.OBJECT_ID, parameters, ocel.object_id_column
    )
    event_id_2 = event_id + "_2"
    object_id_2 = object_id + "_2"

    # Get unique event IDs efficiently
    selected_event_ids = _get_unique_ids(ocel.events, event_id)
    if not selected_event_ids:
        # If no events are selected, clear all components
        for df_attr in ['events', 'objects', 'relations', 'e2e', 'o2o', 'object_changes']:
            setattr(ocel, df_attr, getattr(ocel, df_attr).iloc[0:0])
        return ocel

    # Filter relations and get unique object IDs in one pass
    ocel.relations = ocel.relations[ocel.relations[event_id].isin(selected_event_ids)]
    selected_object_ids = _get_unique_ids(ocel.relations, object_id)

    # Filter remaining components efficiently
    if selected_object_ids:
        # Apply filters to object-related components
        ocel.objects = ocel.objects[ocel.objects[object_id].isin(selected_object_ids)]
        ocel.object_changes = ocel.object_changes[ocel.object_changes[object_id].isin(selected_object_ids)]

        # Combined filtering for o2o with single mask
        o2o_mask = (ocel.o2o[object_id].isin(selected_object_ids) &
                    ocel.o2o[object_id_2].isin(selected_object_ids))
        ocel.o2o = ocel.o2o[o2o_mask]
    else:
        # If no objects are selected, clear objects-related components
        for df_attr in ['objects', 'o2o', 'object_changes']:
            setattr(ocel, df_attr, getattr(ocel, df_attr).iloc[0:0])

    # Filter e2e with single mask
    e2e_mask = (ocel.e2e[event_id].isin(selected_event_ids) &
                ocel.e2e[event_id_2].isin(selected_event_ids))
    ocel.e2e = ocel.e2e[e2e_mask]

    return ocel


def propagate_object_filtering(
        ocel: OCEL, parameters: Optional[Dict[Any, Any]] = None
) -> OCEL:
    """
    Propagates the filtering at the object level to the remaining parts of the OCEL structure
    (events, relations)

    Parameters
    ----------------
    ocel
        Object-centric event log
    parameters
        Parameters of the algorithm, including:
        - Parameters.EVENT_ID => the column to be used as case identifier
        - Parameters.OBJECT_ID => the column to be used as object identifier
        - Parameters.OBJECT_TYPE => the column to be used as object type

    Returns
    ----------------
    ocel
        Object-centric event log with propagated filter
    """
    if parameters is None:
        parameters = {}

    event_id = exec_utils.get_param_value(
        Parameters.EVENT_ID, parameters, ocel.event_id_column
    )
    object_id = exec_utils.get_param_value(
        Parameters.OBJECT_ID, parameters, ocel.object_id_column
    )
    event_id_2 = event_id + "_2"
    object_id_2 = object_id + "_2"

    # Get unique object IDs efficiently
    selected_object_ids = _get_unique_ids(ocel.objects, object_id)
    if not selected_object_ids:
        # If no objects are selected, clear all components
        for df_attr in ['events', 'objects', 'relations', 'e2e', 'o2o', 'object_changes']:
            setattr(ocel, df_attr, getattr(ocel, df_attr).iloc[0:0])
        return ocel

    # Filter relations and get unique event IDs in one pass
    ocel.relations = ocel.relations[ocel.relations[object_id].isin(selected_object_ids)]
    selected_event_ids = _get_unique_ids(ocel.relations, event_id)

    # Filter event-related components
    if selected_event_ids:
        ocel.events = ocel.events[ocel.events[event_id].isin(selected_event_ids)]

        # Combined filtering for e2e with single mask
        e2e_mask = (ocel.e2e[event_id].isin(selected_event_ids) &
                    ocel.e2e[event_id_2].isin(selected_event_ids))
        ocel.e2e = ocel.e2e[e2e_mask]
    else:
        # If no events are selected, clear events-related components
        for df_attr in ['events', 'e2e']:
            setattr(ocel, df_attr, getattr(ocel, df_attr).iloc[0:0])

    # Filter object-related components
    o2o_mask = (ocel.o2o[object_id].isin(selected_object_ids) &
                ocel.o2o[object_id_2].isin(selected_object_ids))
    ocel.o2o = ocel.o2o[o2o_mask]

    ocel.object_changes = ocel.object_changes[ocel.object_changes[object_id].isin(selected_object_ids)]

    return ocel


def propagate_relations_filtering(
        ocel: OCEL, parameters: Optional[Dict[Any, Any]] = None
) -> OCEL:
    """
    Propagates the filtering at the relations level to the remaining parts of the OCEL structure
    (events, objects)

    Parameters
    ----------------
    ocel
        Object-centric event log
    parameters
        Parameters of the algorithm, including:
        - Parameters.EVENT_ID => the column to be used as case identifier
        - Parameters.OBJECT_ID => the column to be used as object identifier
        - Parameters.OBJECT_TYPE => the column to be used as object type

    Returns
    ----------------
    ocel
        Object-centric event log with propagated filter
    """
    if parameters is None:
        parameters = {}

    event_id = exec_utils.get_param_value(
        Parameters.EVENT_ID, parameters, ocel.event_id_column
    )
    object_id = exec_utils.get_param_value(
        Parameters.OBJECT_ID, parameters, ocel.object_id_column
    )
    event_id_2 = event_id + "_2"
    object_id_2 = object_id + "_2"

    # Efficiently get unique IDs from relations
    relation_event_ids = _get_unique_ids(ocel.relations, event_id)
    relation_object_ids = _get_unique_ids(ocel.relations, object_id)

    if not relation_event_ids or not relation_object_ids:
        # If no relations are selected, clear all components
        for df_attr in ['events', 'objects', 'relations', 'e2e', 'o2o', 'object_changes']:
            setattr(ocel, df_attr, getattr(ocel, df_attr).iloc[0:0])
        return ocel

    # Compute intersections with existing components
    existing_event_ids = _get_unique_ids(ocel.events, event_id)
    existing_object_ids = _get_unique_ids(ocel.objects, object_id)

    selected_event_ids = relation_event_ids.intersection(existing_event_ids)
    selected_object_ids = relation_object_ids.intersection(existing_object_ids)

    if not selected_event_ids or not selected_object_ids:
        # If no valid intersections, clear all components
        for df_attr in ['events', 'objects', 'relations', 'e2e', 'o2o', 'object_changes']:
            setattr(ocel, df_attr, getattr(ocel, df_attr).iloc[0:0])
        return ocel

    # Filter all components in a single operation each
    ocel.events = ocel.events[ocel.events[event_id].isin(selected_event_ids)]
    ocel.objects = ocel.objects[ocel.objects[object_id].isin(selected_object_ids)]

    # Filter relations with combined condition
    relations_mask = (ocel.relations[event_id].isin(selected_event_ids) &
                      ocel.relations[object_id].isin(selected_object_ids))
    ocel.relations = ocel.relations[relations_mask]

    # Filter e2e and o2o with combined conditions
    e2e_mask = (ocel.e2e[event_id].isin(selected_event_ids) &
                ocel.e2e[event_id_2].isin(selected_event_ids))
    ocel.e2e = ocel.e2e[e2e_mask]

    o2o_mask = (ocel.o2o[object_id].isin(selected_object_ids) &
                ocel.o2o[object_id_2].isin(selected_object_ids))
    ocel.o2o = ocel.o2o[o2o_mask]

    ocel.object_changes = ocel.object_changes[ocel.object_changes[object_id].isin(selected_object_ids)]

    return ocel
